#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on June 7, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QBoxLayout

# A handy limit for things measured in pixels
SCREEN_MAX_VALUE = 100000

SCENE_BORDER_WIDTH = 10

GRID_STEP = 10

# Karabo colors
HOVER_COLOR = (75, 138, 164)
SELECTION_COLOR = (11, 18, 61)


# For convenience association of strings to Qt cursors
QT_CURSORS = {
    'arrow': Qt.ArrowCursor,
    'blank': Qt.BlankCursor,
    'busy': Qt.BusyCursor,
    'cross': Qt.CrossCursor,
    'closed-hand': Qt.ClosedHandCursor,
    'open-hand': Qt.OpenHandCursor,
    'pointing-hand': Qt.PointingHandCursor,
    'resize-diagonal-trbl': Qt.SizeBDiagCursor,
    'resize-diagonal-tlbr': Qt.SizeFDiagCursor,
    'resize-horizontal': Qt.SizeHorCursor,
    'resize-vertical': Qt.SizeVerCursor,
}

# For convenience association of strings to Qt enums
QT_PEN_CAP_STYLE_FROM_STR = {
    'butt': Qt.FlatCap,
    'square': Qt.SquareCap,
    'round': Qt.RoundCap,
}

# For convenience association reverse mapping of Qt enums to string
QT_PEN_CAP_STYLE_TO_STR = {v: k for k, v in QT_PEN_CAP_STYLE_FROM_STR.items()}

# For convenience association of strings to Qt enums
QT_PEN_JOIN_STYLE_FROM_STR = {
    'miter': Qt.MiterJoin,
    'svgmiter': Qt.SvgMiterJoin,
    'round': Qt.RoundJoin,
    'bevel': Qt.BevelJoin,
}

# For convenience association reverse mapping of Qt enums to string
QT_PEN_JOIN_STYLE_TO_STR = {v: k
                            for k, v in QT_PEN_JOIN_STYLE_FROM_STR.items()}

# For convenience association of ints to Qt enums
QT_BOX_LAYOUT_DIRECTION = (
    QBoxLayout.LeftToRight,
    QBoxLayout.RightToLeft,
    QBoxLayout.TopToBottom,
    QBoxLayout.BottomToTop
)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .axis_item import (
    AlarmAxisItem, AuxPlotAxisItem, AxisItem, StateAxisItem, TimeAxisItem,
    create_axis_items)
from .canvas import PointCanvas, RectCanvas
from .colors import (
    COLORMAPS, Colors, get_available_colors, get_brush_cycler,
    get_default_brush, get_default_pen, get_pen_cycler, make_brush, make_pen,
    rgba_to_hex)
from .const import get_alarm_string, get_state_string
from .dialogs.axes_labels import AxesLabelsDialog
from .enums import (
    AspectRatio, AuxPlots, Axes, AxisType, ExportTool, MouseMode, MouseTool,
    ROITool)
from .exporters import ArrayExporter, ImageExporter, PlotDataExporter
from .legend import CoordsLegend, KaraboLegend
from .roi.base import KaraboROI
from .roi.controllers.base import BaseROIController
from .roi.controllers.image import ImageROIController
from .roi.crosshair import CrosshairROI
from .roi.rect import RectROI
from .roi.utils import ImageRegion
from .toolbar.controller import ToolbarController
from .toolbar.toolsets import BaseToolsetController, ExportToolset, ROIToolset
from .toolbar.widgets import create_button, create_tool_button
from .utils import float_to_string, safe_log10
from .viewbox import KaraboViewBox



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QPoint, QRect
from qtpy.QtWidgets import QDialog, QWidget

from karabo.common.scenemodel.api import (
    DeviceSceneLinkModel, LabelModel, SceneLinkModel, SceneTargetWindow,
    WebLinkModel)
from karabo.native import Hash
from karabogui.events import KaraboEvent
from karabogui.testing import GuiTestCase, click_button, singletons

from ..link import (
    DeviceSceneLinkWidget, SceneLinkWidget, WebDialog, WebLinkWidget)


class TestSceneLink(GuiTestCase):

    def setUp(self):
        super().setUp()
        false_font = "Times New Roman,10,-1,5,50,0,0,0,0,0"
        self.model = SceneLinkModel(
            text="SceneLink Text", foreground='black',
            x=0, y=0, width=100, height=100, font=false_font,
            target="name:122345677")

        self.main_widget = QWidget()
        self.widget = SceneLinkWidget(model=self.model,
                                      parent=self.main_widget)
        self.main_widget.show()

    def tearDown(self):
        super().tearDown()
        self.main_widget = None
        self.widget = None

    # -----------------------------------------------------------------------
    # Actual tests

    def test_font_replacement(self):
        font = self.widget.model.font
        self.assertEqual(font, "Source Sans Pro,10,-1,5,50,0,0,0,0,0")

    def test_basics(self):
        model_rect = QRect(self.model.x, self.model.y,
                           self.model.width, self.model.height, )
        self.assertTrue(self.widget.rect() == model_rect)
        self.assertEqual(len(self.widget.actions()), 2)
        target = 'karabogui.sceneview.widget.link.broadcast_event'
        with mock.patch(target) as broadcast_event:
            self.click(self.widget)
            self.process_qt_events()
            broadcast_event.assert_called_with(
                KaraboEvent.OpenSceneLink,
                {'name': 'name', 'target': '122345677',
                 'target_window': SceneTargetWindow.MainWindow})

        # Make sure, basic interface is there
        self.widget.add_proxies(None)
        self.widget.apply_changes()
        self.widget.decline_changes()
        self.widget.destroy()
        self.widget.set_visible(True)
        self.widget.update_global_access_level(None)
        self.assertIsNotNone(self.widget)

        # Geometry
        rect = QRect(2, 10, 2, 10)
        self.assertFalse(self.widget.geometry() == rect)
        self.widget.set_geometry(rect)
        self.assertTrue(self.widget.geometry() == rect)

        # Translation
        self.assertEqual(self.widget.pos(), QPoint(2, 10))
        self.widget.translate(QPoint(10, 0))
        self.assertEqual(self.widget.pos(), QPoint(12, 10))

        path = "karabogui.sceneview.widget.link.TextDialog"
        with mock.patch(path) as dialog:
            dialog.exec.return_value = QDialog.Accepted
            dialog().label_model = LabelModel(text="This is a label text")
            self.widget.actions()[0].trigger()
            self.assertEqual(self.model.text, "This is a label text")

        path = "karabogui.sceneview.widget.link.SceneLinkDialog"
        with mock.patch(path) as dialog:
            dialog.exec.return_value = QDialog.Accepted
            dialog().selectedScene = "New selected scene"
            dialog().selectedTargetWindow = SceneTargetWindow.Dialog
            self.widget.actions()[1].trigger()
            self.assertEqual(self.model.target, "New selected scene")
            self.assertEqual(self.model.target_window,
                             SceneTargetWindow.Dialog)
            self.assertEqual(self.widget.toolTip(), "name:122345677")


class TestWeblink(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.model = WebLinkModel(
            text="SceneLink Text", foreground='black',
            x=0, y=0, width=100, height=100,
            target="www.xfel.eu")

        self.main_widget = QWidget()
        self.widget = WebLinkWidget(model=self.model,
                                    parent=self.main_widget)
        self.main_widget.show()

    def tearDown(self):
        super().tearDown()
        self.main_widget = None
        self.widget = None

    # -----------------------------------------------------------------------
    # Actual tests

    def test_basics(self):
        model_rect = QRect(self.model.x, self.model.y,
                           self.model.width, self.model.height, )
        self.assertTrue(self.widget.rect() == model_rect)
        self.assertEqual(len(self.widget.actions()), 2)
        target = 'karabogui.sceneview.widget.link.webbrowser'
        with mock.patch(target) as browser:
            self.click(self.widget)
            self.process_qt_events()
            browser.open_new.assert_called_with("www.xfel.eu")

        # Make sure, basic interface is there
        self.widget.add_proxies(None)
        self.widget.apply_changes()
        self.widget.decline_changes()
        self.widget.destroy()
        self.widget.set_visible(True)
        self.widget.update_global_access_level(None)
        self.assertIsNotNone(self.widget)

        # Geometry
        rect = QRect(2, 10, 2, 10)
        self.assertFalse(self.widget.geometry() == rect)
        self.widget.set_geometry(rect)
        self.assertTrue(self.widget.geometry() == rect)

        # Translation
        self.assertEqual(self.widget.pos(), QPoint(2, 10))
        self.widget.translate(QPoint(10, 0))
        self.assertEqual(self.widget.pos(), QPoint(12, 10))

        path = "karabogui.sceneview.widget.link.TextDialog"
        with mock.patch(path) as dialog:
            dialog.exec.return_value = QDialog.Accepted
            dialog().label_model = LabelModel(text="This is a label text")
            self.widget.actions()[0].trigger()
            self.assertEqual(self.model.text, "This is a label text")

        path = "karabogui.sceneview.widget.link.WebDialog"
        with mock.patch(path) as dialog:
            dialog.exec.return_value = QDialog.Accepted
            dialog().target = "www.desy.de"
            self.widget.actions()[1].trigger()
            self.assertEqual(self.model.target, "www.desy.de")
            self.assertEqual(self.widget.toolTip(), "www.xfel.eu")

    def test_webdialog(self):
        dialog = WebDialog(self.model.target)
        self.assertNotEqual(dialog.target, self.model.target)
        # Dialog corrects with http
        self.assertEqual(dialog.target, "http://www.xfel.eu")


def test_device_scene_link_widget(gui_app, mocker):
    model = DeviceSceneLinkModel(
        text="DeviceSceneLink Text", foreground="black",
        x=0, y=0, width=100, height=100,
        target="scene", keys=["XHQ_EH_DG/SCENE/LINK"])

    main_widget = QWidget()
    widget = DeviceSceneLinkWidget(model=model,
                                   parent=main_widget)
    main_widget.show()

    model_rect = QRect(model.x, model.y,
                       model.width, model.height)
    assert widget.rect() == model_rect
    assert len(widget.actions()) == 2

    # Geometry
    rect = QRect(2, 10, 2, 10)
    assert widget.geometry() != rect
    widget.set_geometry(rect)
    assert widget.geometry() == rect

    # Translation
    assert widget.pos() == QPoint(2, 10)
    widget.translate(QPoint(10, 0))
    assert widget.pos() == QPoint(12, 10)

    path = "karabogui.sceneview.widget.link.TextDialog"
    dialog = mocker.patch(path)
    dialog.exec.return_value = QDialog.Accepted
    dialog().label_model = LabelModel(text="This is a label text")
    widget.actions()[0].trigger()
    assert model.text == "This is a label text"

    path = "karabogui.sceneview.widget.link.DeviceCapabilityDialog"
    dialog = mocker.patch(path)
    dialog.exec.return_value = QDialog.Accepted
    dialog().capa_name = "plot"
    widget.actions()[1].trigger()
    assert model.target == "plot"
    assert widget.toolTip() == "XHQ_EH_DG/SCENE/LINK|scene"

    # Make sure, basic interface is there
    widget.add_proxies(None)
    widget.apply_changes()
    widget.decline_changes()
    widget.set_visible(True)
    widget.update_global_access_level(None)
    widget.destroy()


def test_device_scene_link_sending(gui_app, mocker):
    # Send a scene target
    model = DeviceSceneLinkModel(
        text="DeviceSceneLink Text", foreground="black",
        x=0, y=0, width=100, height=100,
        target="scene", keys=["XHQ_EH_DG/SCENE/LINK"])

    main_widget = QWidget()
    widget = DeviceSceneLinkWidget(model=model,
                                   parent=main_widget)
    main_widget.show()
    manager = mocker.Mock()
    path = "karabogui.sceneview.widget.link.is_device_online"
    question = mocker.patch(path)
    question.return_value = True
    with singletons(manager=manager):
        click_button(widget)
        args = manager.callDeviceSlot.call_args
        assert len(args) == 2
        assert args[0][1] == "XHQ_EH_DG/SCENE/LINK"
        assert args[0][2] == "requestScene"
        # of args the last argument is the parameter hash
        h = args[0][-1]
        assert isinstance(h, Hash)
        assert "name" in h
        assert h["name"] == "scene"

        # Erase target and click again, empty string is send
        model.target = ""
        click_button(widget)
        args = manager.callDeviceSlot.call_args
        h = args[0][-1]
        assert isinstance(h, Hash)
        assert "name" in h
        assert h["name"] == ""

    widget.destroy()


if __name__ == "__main__":
    main()



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 27, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from abc import abstractmethod

from qtpy.QtCore import QAbstractItemModel, Qt
from qtpy.QtGui import QBrush, QFont, QIcon
from traits.api import (
    ABCHasStrictTraits, Callable, Dict, Enum, HasStrictTraits, Instance, List,
    Property, String, WeakRef, on_trait_change)

from karabo.common.api import InstanceStatus
from karabo.common.project.api import BaseProjectObjectModel
from karabogui.binding.api import ProxyStatus


class ProjectControllerUiData(HasStrictTraits):
    """ A data class which contains all necessary data for the Qt item to
    display
    """
    icon = Instance(QIcon, args=())
    font = Instance(QFont, args=())
    brush = Instance(QBrush, args=())
    status = Enum(*ProxyStatus)
    instance_status = Enum(*InstanceStatus)


class BaseProjectController(ABCHasStrictTraits):
    """ A base class for all objects which control ProjectModel objects for use
    in the ProjectTreeView.
    """
    # The project model object controlled here
    model = Instance(BaseProjectObjectModel)

    # A back-reference to our parent controller
    parent = WeakRef('BaseProjectController', allow_none=True)

    # Name to be displayed in the item view
    display_name = Property(String)

    # Easy access display data for model item
    ui_data = Instance(ProjectControllerUiData)

    # Reference the QAbstractItemModel
    _qt_model = WeakRef(QAbstractItemModel)

    @abstractmethod
    def context_menu(self, project_controller, parent=None):
        """ Requests a menu to be used as a context menu for this item.

        :param project_controller: The ProjectController which is the immediate
                                   parent of the item which was clicked on.
        :param parent: A QObject which can be passed as a Qt object parent.
        :return: A QMenu containing contextual actions for the item.
        """

    @abstractmethod
    def create_ui_data(self):
        """Must be implemented by derived classes to return a
        `ProjectControllerUiData` instance representing this controller. The
        result will be stored in the ``ui_data`` trait.
        """

    def double_click(self, project_controller, parent=None):
        """ Handles a double click event on this item.

        :param project_controller: The ProjectController which is the immediate
                                   parent of the item which was clicked on.
        :param parent: A QObject which can be passed as a Qt object parent.
        """

    def single_click(self, project_controller, parent=None):
        """ Handles a single click event on this item.

        :param project_controller: The ProjectController which is the immediate
                                   parent of the item which was clicked on.
        :param parent: A QObject which can be passed as a Qt object parent.
        """

    def delete_press(self, project_controller, parent=None):
        """ Handles a delete key press event on this item.

        :param project_controller: The ProjectController which is the immediate
                                   parent of the item which was pressed on.
        :param parent: A QObject which can be passed as a Qt object parent.
        """

    def child(self, index):
        """Returns a child of this controller.

        :param index: An index into the list of this controller's children
        :returns: A BaseProjectController instance or None
        """
        return None

    def rows(self):
        """Returns the number of rows 'under' this controller in the project
        tree view.
        """
        return -1

    def ui_item_text(self):
        """The name for this controller's item in the GUI view
        """
        if self.model.modified:
            return f'*{self.display_name}'
        else:
            return self.display_name

    @on_trait_change('model.modified')
    def _update_label_style(self):
        brush = QBrush(self.ui_data.brush)
        if self.model.modified:
            brush.setColor(Qt.blue)
        else:
            brush.setColor(Qt.black)
        self.ui_data.brush = brush

    @on_trait_change('model.simple_name')
    def _update_label(self):
        """ Whenever ``simple_name`` is modified the UI should repaint
        """
        if self._qt_model is not None:
            self._qt_model.controller_data_update(self, roles=[Qt.DisplayRole])

    @on_trait_change('ui_data.icon,ui_data.brush')
    def _request_repaint(self):
        if self._qt_model is not None:
            self._qt_model.controller_data_update(
                self, roles=[Qt.DecorationRole, Qt.FontRole,
                             Qt.ForegroundRole])

    @on_trait_change('ui_data.instance_status')
    def _update_instance_info(self):
        if self._qt_model is not None:
            self._qt_model.controller_data_update(
                self, column=1, roles=[Qt.DecorationRole])

    def _get_display_name(self):
        """Traits property getter for ``display_name``
        """
        return self.model.simple_name

    def _ui_data_default(self):
        """Traits default initializer for `ui_data`
        """
        return self.create_ui_data()


class BaseProjectGroupController(BaseProjectController):
    """ A base class for all project controller objects which have children.
    """
    # A factory for children controllers
    child_create = Callable
    # A callable which can gracefully destroy a child controller
    child_destroy = Callable
    # Child controllers
    children = List(Instance(BaseProjectController))
    # Dictionary for fast lookups during removal
    _child_map = Dict  # {model: controller}

    def child(self, index):
        """Return child at given ``index``
        """
        return self.children[index]

    def rows(self):
        """Return number of rows
        """
        return len(self.children)

    def items_assigned(self, obj, name, old, new):
        """ Handles assignment to a list trait and passes the notification
        along to ``item_handler``
        """
        self.item_handler(new, old)

    def items_mutated(self, event):
        """ Handles mutation(insertion/removal) of a list trait and passes the
        notification along to ``item_handler``
        """
        self.item_handler(event.added, event.removed)

    def item_handler(self, added, removed):
        """ Called for List-trait events on ``model``
        """
        for model in removed:
            controller = self._child_map[model]
            # NOTE: Always use a context manager when modifying self.children!
            with self._qt_model.removal_context(controller):
                self.children.remove(controller)
            self.child_destroy(controller)

        # Synchronize the GUI with the Traits model
        additions = [self.child_create(model=model, parent=self,
                                       _qt_model=self._qt_model)
                     for model in added]
        self._update_ui_children(additions)

    def _children_items_changed(self, event):
        """ Maintain ``_child_map`` by watching item events on ``children``

        This is a static notification handler which is connected automatically
        by Traits.
        """
        for controller in event.removed:
            del self._child_map[controller.model]

        for controller in event.added:
            self._child_map[controller.model] = controller

    def _update_ui_children(self, additions):
        """ Propagate changes from the Traits model to the Qt item model.
        """
        if not additions:
            return

        # First and last indices, INCLUSIVE (thus the '- 1')
        first = len(self.children)
        last = first + len(additions) - 1
        # NOTE: Always use a context manager when modifying self.children!
        with self._qt_model.insertion_context(self, first, last):
            self.children.extend(additions)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on June 10, 2020
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import QLocale, QSize, Qt
from qtpy.QtGui import QFontMetrics
from qtpy.QtSvg import QSvgWidget
from qtpy.QtWidgets import QDoubleSpinBox, QFrame, QLabel, QLineEdit, QSpinBox

from karabogui.const import WIDGET_MIN_HEIGHT, WIDGET_MIN_WIDTH
from karabogui.fonts import substitute_font

DEFAULT_SIZE_HINT = 20
CONTENT_MARGIN = 10

SVG_MAX_SIZE = 24

STATE_MIN_SIZE = 10
STATE_WIDTH_HINT = 30

LOCALE = QLocale("en_US")


class KaraboSceneWidget:

    def __init__(self, *args, model=None, **kwargs):
        """Check and substitute the font with the application fonts"""
        self.model = model
        substitute_font(model)
        super().__init__(*args, **kwargs)

    def sizeHint(self):
        """On old GUI, the geometry of the widgets in a layout are saved
        with smaller dimensions and with (0, 0) position. We only use the
        model geometry as size hint if it is the corrected one."""
        size = QSize()
        if self._has_model_pos():
            size = QSize(self.model.width, self.model.height)
        if size.isEmpty():
            size = super().sizeHint()
        return size

    def minimumSizeHint(self):
        """On old GUI, the geometry of the widgets in a layout are saved
        with smaller dimensions and with (0, 0) position. We only use the
        model geometry as size hint if it is the corrected one."""
        size = QSize()
        if self._has_model_pos():
            size = QSize(self.model.width, self.model.height)
        if size.isEmpty():
            size = super().minimumSizeHint()
        return size

    def _has_model_pos(self):
        model = self.model
        return model is not None and (model.x, model.y) != (0, 0)


class Label(QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumWidth(WIDGET_MIN_WIDTH)
        self.setMinimumHeight(WIDGET_MIN_HEIGHT)
        self.setAlignment(Qt.AlignCenter)
        self.setWordWrap(True)

    def sizeHint(self):
        fm = QFontMetrics(self.font())
        width = fm.width(self.text()) + CONTENT_MARGIN

        return QSize(width, DEFAULT_SIZE_HINT)


class ElidingLabel(QLabel):
    """A label that elides if the text and margins are larger than the width
    """

    def __init__(self, parent=None):
        self._text = ""
        self._elided = False
        super().__init__(self._text, parent=parent)
        self.setMinimumWidth(WIDGET_MIN_WIDTH)
        self.setMinimumHeight(WIDGET_MIN_HEIGHT)
        self.setAlignment(Qt.AlignCenter)

    def isElided(self):
        """Return if the text is elided or not"""
        return self._elided

    def sizeHint(self):
        """Reimplemented function of `QLabel` for the size hint calculation"""
        fm = QFontMetrics(self.font())
        width = fm.width(self.text()) + CONTENT_MARGIN
        return QSize(width, WIDGET_MIN_HEIGHT)

    @property
    def text_width(self):
        return self.width() - CONTENT_MARGIN

    def setText(self, text):
        """Reimplemented function of `QLabel` to adjust the display text"""
        self._text = text
        text_width = self.text_width
        fm = QFontMetrics(self.font())
        elided = False
        if fm.width(text) > text_width:
            text = fm.elidedText(text, Qt.ElideRight, text_width)
            elided = True
        self._elided = elided
        if text != self.text():
            super().setText(text)

    def resizeEvent(self, event):
        """Reimplemented function of `QLabel` to adjust the display text"""
        super().resizeEvent(event)
        self.setText(self._text)


class LineEdit(QLineEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumWidth(WIDGET_MIN_WIDTH)
        self.setMinimumHeight(WIDGET_MIN_HEIGHT)
        self.setAlignment(Qt.AlignLeft | Qt.AlignAbsolute)

    def sizeHint(self):
        fm = QFontMetrics(self.font())
        width = fm.width(self.text()) + CONTENT_MARGIN

        return QSize(width, DEFAULT_SIZE_HINT)


class SpinBox(QSpinBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumWidth(WIDGET_MIN_WIDTH)
        self.setMinimumHeight(WIDGET_MIN_HEIGHT)
        self.setAlignment(Qt.AlignLeft | Qt.AlignAbsolute)
        self.setLocale(LOCALE)
        self.setFocusPolicy(Qt.StrongFocus)

    def sizeHint(self):
        fm = QFontMetrics(self.font())
        width = fm.width(self.cleanText()) + CONTENT_MARGIN
        return QSize(width, DEFAULT_SIZE_HINT)


class DoubleSpinBox(QDoubleSpinBox):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMinimumWidth(WIDGET_MIN_WIDTH)
        self.setMinimumHeight(WIDGET_MIN_HEIGHT)
        self.setAlignment(Qt.AlignLeft | Qt.AlignAbsolute)
        self.setLocale(LOCALE)
        self.setFocusPolicy(Qt.StrongFocus)

    def sizeHint(self):
        fm = QFontMetrics(self.font())
        width = fm.width(self.cleanText()) + CONTENT_MARGIN
        return QSize(width, DEFAULT_SIZE_HINT)


class SvgWidget(QSvgWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setMaximumSize(SVG_MAX_SIZE, SVG_MAX_SIZE)

    def sizeHint(self):
        return QSize(DEFAULT_SIZE_HINT, DEFAULT_SIZE_HINT)


class FrameWidget(QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAutoFillBackground(True)
        self.setAlignment(Qt.AlignCenter)
        self.setMinimumWidth(STATE_MIN_SIZE)
        self.setMinimumHeight(STATE_MIN_SIZE)
        self.setWordWrap(True)
        self.setFrameStyle(QFrame.Box | QFrame.Plain)

    def sizeHint(self):
        return QSize(STATE_WIDTH_HINT, DEFAULT_SIZE_HINT)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os.path as op

from karabogui import icons

from .macropanel import MacroPanel
from .scenepanel import ScenePanel

PANEL_ICONS = {
    MacroPanel: icons.edit,
    ScenePanel: icons.image
}


def get_panel_icon(panel):
    return PANEL_ICONS.get(type(panel))


def format_vector_hash_details(binding, value):
    if value is None:
        return "Schema is not followed"

    details = []
    for index, row_hash in enumerate(value):
        if row_hash is None:
            continue
        row_details = []
        for row_name, row_value in row_hash.items():
            row_binding = binding.bindings.get(row_name)
            detail = format_property_details(row_binding, row_name, row_value)
            row_details.append(detail)
        # Do some formatting, a la yaml.
        tabbed_details = [f"    {detail}" for detail in row_details]
        tabbed_row = [f"    {deets}"
                      for deets in [f"row {index}:"] + tabbed_details]
        details.append("\n".join(tabbed_row))
    return "\n" + "\n".join(details)


def format_property_details(binding, path, value=''):
    name = binding.displayedName or path
    data_type = type(binding).__name__[:-len('binding')]
    return f"- {name} ({data_type}): {value}"


def compare_proxy_essential(old, new):
    """Compare the necessary proxy information and return True if equal

    Note: Proxies might change their classes, e.g. are different from project
    and online information. Hence, they are not considered.

    :returns: Boolean if essentials are equal (serverId, deviceId)
    """

    def get_proxy_id(proxy):
        server_id = proxy.server_id
        device_id = getattr(proxy, "device_id", "")
        return tuple([server_id, device_id])

    return get_proxy_id(old) == get_proxy_id(new)


def get_panel_ui(ui_name):
    assert ui_name.endswith(".ui"), f"{ui_name} is not a `.ui` file"

    return op.join(op.abspath(op.dirname(__file__)), "ui", ui_name)



import json
from ast import literal_eval
from datetime import datetime, timezone

from qtpy.QtCore import Qt
from qtpy.QtNetwork import QNetworkReply, QNetworkRequest

from karabogui import access as krb_access
from karabogui.const import IS_WINDOWS_SYSTEM
from karabogui.dialogs.reactive_login_dialog import (
    AccessCodeWidget, LoginType, ReactiveLoginDialog, UserSessionDialog,
    remaining_time_info)
from karabogui.singletons.configuration import Configuration
from karabogui.testing.utils import click_button, keySequence, singletons


def test_access_level(gui_app):
    configuration = Configuration()

    with singletons(configuration=configuration):
        dialog = ReactiveLoginDialog()
        assert dialog.access_level == (
            krb_access.GLOBAL_ACCESS_LEVEL.name.lower())

        # Non-authenticated login type, access level from the dialog.
        dialog.login_type = LoginType.ACCESS_LEVEL
        assert dialog.combo_access_level.currentText() == "expert"
        assert dialog.access_level == "expert"

        # For authenticated login type, access level from configuration.
        configuration["access_level"] = "operator"
        dialog.login_type = LoginType.REFRESH_TOKEN
        assert dialog.access_level == "operator"


def test_temporarySessionDialog(gui_app, mocker):
    dialog = UserSessionDialog()
    assert not dialog.ok_button.isEnabled()
    access_code = "123456"
    hostname = "karabo.xfel.eu"
    for i, cell in enumerate(dialog.edit_access_code.cells, start=1):
        cell.setText(str(i))
    mocker.patch("karabogui.dialogs.reactive_login_dialog.CLIENT_HOST",
                 new=hostname)
    dialog.access_manager = mocker.Mock()
    dialog._login_authenticated()
    assert dialog.access_manager.post.call_count == 1
    call_args, _ = dialog.access_manager.post.call_args
    network_req, args = call_args
    assert isinstance(network_req, QNetworkRequest)
    args_dict = json.loads(args)
    assert args_dict["access_code"] == int(access_code)
    assert not args_dict["remember_login"]
    assert args_dict["client_hostname"] == hostname


def test_switch_user(gui_app, mocker):
    krb_access.SESSION_END_NOTICE = True
    dialog = UserSessionDialog()
    assert not dialog.ok_button.isEnabled()
    access_code = "123456"
    hostname = "karabo.xfel.eu"
    for i, cell in enumerate(dialog.edit_access_code.cells, start=1):
        cell.setText(str(i))
    mocker.patch("karabogui.dialogs.reactive_login_dialog.CLIENT_HOST",
                 new=hostname)
    dialog.access_manager = mocker.Mock()
    assert dialog.combo_mode.currentIndex() == 1
    dialog.remember_login.setChecked(True)
    dialog._login_authenticated()
    assert dialog.access_manager.post.call_count == 1
    call_args, _ = dialog.access_manager.post.call_args
    network_req, args = call_args
    assert isinstance(network_req, QNetworkRequest)
    args_dict = json.loads(args)
    assert args_dict["access_code"] == int(access_code)
    assert args_dict["remember_login"]
    assert args_dict["client_hostname"] == hostname

    dialog.remember_login.setChecked(False)
    dialog._login_authenticated()
    assert dialog.access_manager.post.call_count == 2
    call_args, _ = dialog.access_manager.post.call_args
    network_req, args = call_args
    assert isinstance(network_req, QNetworkRequest)
    args_dict = json.loads(args)
    assert args_dict["access_code"] == int(access_code)
    assert not args_dict["remember_login"]
    assert args_dict["client_hostname"] == hostname
    krb_access.SESSION_END_NOTICE = False


def test_connect_button(gui_app):
    dialog = ReactiveLoginDialog()
    assert not dialog.connect_button.isEnabled()
    dialog.login_type = LoginType.REFRESH_TOKEN
    dialog._update_button()
    assert dialog.connect_button.isEnabled()

    dialog._on_switch_user_click()
    assert not dialog.connect_button.isEnabled()


def test_refresh_token(gui_app, mocker):
    configuration = Configuration()
    with singletons(configuration=configuration):
        configuration["refresh_token"] = "dummy_token"
        dialog = ReactiveLoginDialog()
        dialog.login_type = LoginType.REFRESH_TOKEN
        mocked_method = mocker.patch.object(
            dialog, "_refresh_authentication")
        dialog.connect_clicked()
        assert mocked_method.call_count


def test_refresh_authentication(gui_app, mocker):
    """With REFRESH_TOKEN, the login dialog should call the correct post
    request."""
    configuration = Configuration()
    with singletons(configuration=configuration):
        configuration["refresh_token"] = "dummy_token"
        configuration["refresh_token_user"] = "karabo"
        dialog = ReactiveLoginDialog()
        dialog.login_type = LoginType.REFRESH_TOKEN
        mocked_access_manger = mocker.patch.object(dialog, "access_manager")
        dialog._refresh_authentication()
        assert mocked_access_manger.post.call_count == 1

        args = mocked_access_manger.post.call_args[0]
        network, args = args

        assert isinstance(network, QNetworkRequest)
        url = network.url().url()
        assert "refresh_tokens" in url

        args = literal_eval(args.decode())
        assert len(args) == 3
        assert args.get("refresh_token") == "dummy_token"


def test_access_code_login(gui_app, mocker):
    """Verify the correct request is posted when loging with an access code
    even when refresh token is already present."""
    configuration = Configuration()
    with singletons(configuration=configuration):
        configuration["refresh_token"] = "dummy_token"
        dialog = ReactiveLoginDialog()
        dialog.login_type = LoginType.REFRESH_TOKEN
        dialog._update_button()
        assert dialog.connect_button.isEnabled()

        click_button(dialog.switch_button)
        assert dialog.login_type == LoginType.USER_AUTHENTICATED
        assert not dialog.connect_button.isEnabled()

        for i, cell in enumerate(dialog.edit_access_code.cells, start=1):
            cell.setText(str(i))

        dialog._update_button()
        assert dialog.connect_button.isEnabled()
        post_access_code = mocker.patch.object(dialog, "_login_authenticated")
        post_refresh_token = mocker.patch.object(
            dialog, "_refresh_authentication")
        click_button(dialog.connect_button)
        assert post_refresh_token.call_count == 0
        assert post_access_code.call_count == 1


def test_authReply(gui_app, mocker):
    configuration = Configuration()
    with singletons(configuration=configuration):
        dialog = ReactiveLoginDialog()
        reply = mocker.Mock(spec=QNetworkReply)
        reply.error.return_value = QNetworkReply.NoError

        failure = (
            b'{"success":false,"once_token":null,"refresh_token":null,'
            b'"username":null,"error_msg":"No token for access code 444444"}')
        reply.readAll.return_value = failure
        dialog.onAuthReply(reply)
        assert dialog.label_status.text() == "No token for access code 444444"

        success = (b'{"success":true,"once_token":"abc","refresh_token":"xyz",'
                   b'"username":"karabo","error_msg":null}')
        reply.readAll.return_value = success
        dialog.onAuthReply(reply)
        assert krb_access.ONE_TIME_TOKEN == "abc"
        assert configuration["refresh_token"] == "xyz"


def test_access_widget(gui_app):
    widget = AccessCodeWidget()
    assert widget.get_access_code() == ""
    assert widget.objectName() == "AccessCodeWidget"
    for i, cell in enumerate(widget.cells):
        assert cell.objectName() == f"Cell_{i}"
    widget.cells[0].setText("4")
    assert widget.get_access_code() == "4"
    assert widget.focusWidget() == widget.cells[1]

    widget.cells[3].setText("123456")
    assert widget.get_access_code() == "123456"
    assert widget.cells[0].text() == "1"
    assert widget.cells[1].text() == "2"
    assert widget.cells[2].text() == "3"
    assert widget.cells[3].text() == "4"
    assert widget.cells[4].text() == "5"
    assert widget.cells[5].text() == "6"

    # Backspace should clear the cell
    cell = widget.focusWidget()
    current_index = widget.cells.index(cell)
    assert current_index
    assert cell.text()
    keySequence(cell, Qt.Key_Backspace)
    new_cell = widget.focusWidget()
    new_index = widget.cells.index(new_cell)
    assert not cell.text()
    assert new_index == current_index

    # Backspace should move focus to previous cell from empty cell.
    keySequence(cell, Qt.Key_Backspace)
    new_cell = widget.focusWidget()
    new_index = widget.cells.index(new_cell)
    assert new_index == current_index - 1

    # Backspace when focus is on first cell.
    first_cell = widget.cells[0]
    first_cell.setFocus(True)
    keySequence(first_cell, Qt.Key_Backspace)
    assert first_cell == widget.focusWidget()

    cells = widget.cells
    for _ in range(7):
        cell = widget.focusWidget()
        cell_index = cells.index(cell)
        keySequence(cell, Qt.Key_Right)
        if cell_index < len(cells) - 1:
            assert cells[cell_index + 1] == widget.focusWidget()
        else:
            assert cells[cell_index] == widget.focusWidget()

    for _ in range(7):
        cell = widget.focusWidget()
        cell_index = cells.index(cell)
        keySequence(cell, Qt.Key_Left)
        if cell_index == 0:
            assert cells[cell_index] == widget.focusWidget()
        else:
            assert cells[cell_index - 1] == widget.focusWidget()

    # Backspace deletes the number when cursor is on its right.
    fourth_cell = cells[4]
    third_cell = cells[3]
    fourth_cell.setFocus(True)
    fourth_cell.setCursorPosition(1)
    keySequence(fourth_cell, Qt.Key_Backspace)
    assert not bool(fourth_cell.text())
    assert fourth_cell == widget.focusWidget()

    # Backspace switches to previous cell if the cursor is on left.
    third_cell.setFocus(True)
    third_cell.setCursorPosition(0)
    keySequence(third_cell, Qt.Key_Backspace)
    assert bool(third_cell.text())
    assert cells[2] == widget.focusWidget()


def test_session_dialog(mocker):
    """Test the session dialog and the remaining time"""
    def create_start_str() -> str:
        """Generates a UTC datetime string in expected format"""
        now_utc = datetime.now(timezone.utc)
        return now_utc.strftime("%Y%m%dT%H%M%S.%fZ")

    # 30 minutes from now
    start = create_start_str()
    duration = 30 * 60  # 30 minutes
    _, _, remaining = remaining_time_info(start, duration)
    if IS_WINDOWS_SYSTEM:
        assert "~30 minute(s) left" in remaining
    else:
        assert "~29 minute(s) left" in remaining

    # 2 hours from now
    start = create_start_str()
    duration = 2 * 3600
    _, _, remaining = remaining_time_info(start, duration)
    assert "~2.00 hour(s)" in remaining

    # 3 days from now
    start = create_start_str()
    duration = 3 * 86400
    _, _, remaining = remaining_time_info(start, duration)
    assert "~3.00 day" in remaining

    # 1 hours ago
    start = create_start_str()
    duration = -3600
    _, _, remaining = remaining_time_info(start, duration)
    assert remaining == "Session ended"

    # test the dialog
    network = mocker.Mock()
    with singletons(network=network):
        dialog = UserSessionDialog()
        network.onGetGuiSessionInfo.assert_called()
        start_time = create_start_str()
        duration = 2 * 3600
        data = {"sessionStartTime": start_time, "sessionDuration": duration}
        dialog._event_user_session_info(data)
        assert dialog.info_remaining_time.text() == "~2.00 hour(s) left"



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on October 9, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op

from qtpy.QtSvg import QSvgWidget
from traits.api import Instance

from karabo.common.scenemodel.api import CheckBoxModel
from karabogui import icons
from karabogui.binding.api import BoolBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)

ICONS = op.dirname(icons.__file__)
CHECKED = op.join(ICONS, "checkbox-checked.svg")
UNCHECKED = op.join(ICONS, "checkbox-unchecked.svg")


@register_binding_controller(ui_name='Toggle Field',
                             klassname='DisplayCheckBox',
                             binding_type=BoolBinding, priority=10)
class DisplayCheckBox(BaseBindingController):
    # The scene data model class for this controller
    model = Instance(CheckBoxModel, args=())

    def create_widget(self, parent):
        widget = QSvgWidget(parent)
        widget.setFixedSize(20, 20)
        return widget

    def value_update(self, proxy):
        svg_file = CHECKED if proxy.value else UNCHECKED
        self.widget.load(svg_file)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtGui import QFont, QFontDatabase

from karabogui.fonts import FONT_FILENAMES, get_alias_from_font
from karabogui.testing import GuiTestCase

from ..font_dialog import FontDialog


class TestFontDialog(GuiTestCase):

    def setUp(self):
        super().setUp()
        # Add fonts
        for font_file in FONT_FILENAMES:
            QFontDatabase.addApplicationFont(font_file)

        self.qfont = qfont = QFont("Source Sans Pro")
        qfont.setPointSize(20)
        qfont.setBold(True)
        qfont.setItalic(True)
        qfont.setStrikeOut(True)
        qfont.setUnderline(True)

        self.dialog = FontDialog(self.qfont)

    def tearDown(self):
        self.dialog.destroy()

    def test_basics(self):
        self._assert_qfont(self.qfont)

    def test_font_change(self):
        family = "Source Serif Pro"
        qfont = QFont(self.qfont)
        qfont.setFamily(family)

        alias = get_alias_from_font(family)
        self.dialog.font_combobox.setCurrentText(alias)
        self._assert_qfont(qfont)

    def test_font_size_change(self):
        size = 30
        qfont = QFont(self.qfont)
        qfont.setPointSize(size)

        self.dialog.font_size_combobox.setCurrentText(str(size))
        self._assert_qfont(qfont)

    def test_bold_change(self):
        enabled = not self.qfont.bold()
        qfont = QFont(self.qfont)
        qfont.setBold(enabled)

        self.dialog.bold_checkbox.setChecked(enabled)
        self._assert_qfont(qfont)

    def test_italic_change(self):
        enabled = not self.qfont.italic()
        qfont = QFont(self.qfont)
        qfont.setItalic(enabled)

        self.dialog.italic_checkbox.setChecked(enabled)
        self._assert_qfont(qfont)

    def test_strikeout_change(self):
        enabled = not self.qfont.strikeOut()
        qfont = QFont(self.qfont)
        qfont.setStrikeOut(enabled)

        self.dialog.strikeout_checkbox.setChecked(enabled)
        self._assert_qfont(qfont)

    def test_underline_change(self):
        enabled = not self.qfont.underline()
        qfont = QFont(self.qfont)
        qfont.setUnderline(enabled)

        self.dialog.underline_checkbox.setChecked(enabled)
        self._assert_qfont(qfont)

    def _assert_qfont(self, qfont):
        dialog = self.dialog
        assert dialog.qfont == qfont
        assert dialog.font_combobox.currentText() == get_alias_from_font(
            qfont.family())
        assert dialog.font_size_combobox.currentText() == str(
            qfont.pointSize())
        assert dialog.bold_checkbox.isChecked() == qfont.bold()
        assert dialog.italic_checkbox.isChecked() == qfont.italic()
        assert dialog.strikeout_checkbox.isChecked() == qfont.strikeOut()
        assert dialog.underline_checkbox.isChecked() == qfont.underline()

        # Check if font string is equal to the font
        new_font = QFont()
        new_font.fromString(qfont.toString())
        assert dialog.qfont == new_font



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os
import sys
import warnings
from pathlib import Path
from platform import system
from traceback import format_exception, print_exception

from pyqtgraph import setConfigOptions
from qtpy.QtCore import QLocale, Qt
from qtpy.QtGui import QFont, QFontDatabase, QPixmap
from qtpy.QtWidgets import QApplication, QSplashScreen, QStyleFactory

from karabo.common.scenemodel.api import (
    SCENE_DEFAULT_DPI, SCENE_FONT_FAMILY, SCENE_FONT_SIZE)
from karabogui.background import create_background_timer
from karabogui.controllers.api import populate_controller_registry
from karabogui.fonts import FONT_FILENAMES, get_font_size_from_dpi
from karabogui.singletons.api import (
    get_config, get_manager, get_panel_wrangler)
from karabogui.util import get_application_icon, process_qt_events, send_info


def excepthook(exc_type, value, traceback):
    print_exception(exc_type, value, traceback)
    text = "".join(format_exception(exc_type, value, traceback))
    try:
        send_info(type="error", traceback=text)
    except Exception:
        print("Could not send exception to network")


def set_app_info(app, company, url, name):
    """Set the information on the app for QSettings"""
    app.setOrganizationName(company)
    app.setOrganizationDomain(url)
    app.setApplicationName(name)


def create_gui_app(args):
    """Create the QApplication with all necessary fonts and settings"""
    if system() == 'Darwin' and 'QT_MAC_WANTS_LAYER' not in os.environ:
        os.environ['QT_MAC_WANTS_LAYER'] = '1'
        # PyQt 5.12 onwards provides problems for MacOS11.
        # https://github.com/conda-forge/pyqt-feedstock/issues/98
    app = QApplication.instance()
    if app is None:
        app = QApplication(args)
    # Set directly the QSettings environment to have access
    set_app_info(app, "XFEL", "xfel.eu", "KaraboGUI")
    # We check our `KARABO_TEST_GUI` variable before due to Squish cracks!
    if get_config()['highDPI'] and not os.environ.get("KARABO_TEST_GUI"):
        # Create a preliminary QApplication to check system/screen properties.
        # This is needed as setting QApplication attributes should be done
        # before the instantiation (Qt bug as of 5.9).
        # Note: Must take the int of dpi! This is fixed in Qt 5.15 and always
        # active in Qt 6!
        dpi = int(app.primaryScreen().logicalDotsPerInch())
        app.quit()
        del app
        # Set the QApplication attributes before its instantiation.
        # Only apply high DPI scaling when the logical DPI is greater than
        # the default DPI (96). This is usually observed on scaled desktops
        # (e.g., 150% scaling on Windows)
        if dpi > SCENE_DEFAULT_DPI:
            QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
            QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)

        app = QApplication(args)
        # Again set the QSettings environment
        app.setOrganizationName('XFEL')
        app.setOrganizationDomain('xfel.eu')
        app.setApplicationName('KaraboGUI')

    create_background_timer()
    style = QStyleFactory.create("Fusion")
    palette = style.standardPalette()
    app.setStyle(style)
    app.setPalette(palette)

    # Add fonts
    for font_file in FONT_FILENAMES:
        QFontDatabase.addApplicationFont(font_file)

    # Set default application font
    font_size = get_font_size_from_dpi(SCENE_FONT_SIZE)
    font = QFont()
    font.setFamily(SCENE_FONT_FAMILY)
    font.setPointSize(font_size)
    families = QFontDatabase().families()
    if "Ubuntu" in families:
        font.insertSubstitution("Ubuntu", SCENE_FONT_FAMILY)
    if "Sans Serif" in families:
        font.insertSubstitution("Sans Serif", SCENE_FONT_FAMILY)
    app.setFont(font)
    app.setStyleSheet("QPushButton { text-align: left; padding: 5px; }")
    app.setStyleSheet("QToolBar { border: 0px }")
    # Also set the font of the QTreeView and QTableView as it defaults to
    # MS Shell Dlg in Windows for elided text when the QApplication stylesheet
    # has been changed. Any unknown font will do to default to QApp font.
    # [QTBUG-29232]
    app.setStyleSheet(
        f"QTreeView {{ font: {font_size}pt '{SCENE_FONT_FAMILY}'; }} "
        f"QTableView {{ font: {font_size}pt '{SCENE_FONT_FAMILY}'; }}")
    app.setAttribute(Qt.AA_DontShowIconsInMenus, False)

    # set a nice app logo
    icon = get_application_icon()
    app.setWindowIcon(icon)

    QLocale.setDefault(QLocale(QLocale.English, QLocale.UnitedStates))

    # Make sure our applications do not terminate on unhandled python
    # exceptions.
    sys.excepthook = excepthook

    return app


def init_gui(app, use_splash=True):
    """Initialize the GUI.

    Imports are being done inside this function to avoid delaying the display
    of the splash screen. We want the user to know that something is happening
    soon after they launch the GUI.
    """
    # Do some event processing!
    process_qt_events(app)

    if use_splash:
        splash_path = str(Path(__file__).parent / '..' / "icons"
                          / "splash.png")
        splash_img = QPixmap(splash_path)
        splash = QSplashScreen(splash_img, Qt.WindowStaysOnTopHint)
        splash.setMask(splash_img.mask())
        splash.show()
        # NOTE: This is needed to make the splash screen show up...
        splash.showMessage(" ")

    # Do some event processing!
    process_qt_events(app)

    # Start some heavy importing!
    import numpy

    from karabogui import icons

    numpy.set_printoptions(suppress=True, threshold=10)
    setConfigOptions(background=None, foreground="k")

    # Suppress some warnings
    # 1. From scipy.ndimage.zoom
    warnings.filterwarnings('ignore', '.*output shape of zoom.*')
    # 2. From scipy.optimize
    warnings.filterwarnings('ignore', '.*Covariance of the parameters '
                                      'could not be estimated.*')

    # Load the icons
    icons.init()
    # Load the sceneview widget controllers
    populate_controller_registry()
    # Initialize the Manager singleton
    get_manager()
    # Initialize the PanelWrangler and attach the splash screen
    panel_wranger = get_panel_wrangler()
    process_qt_events(app)
    if use_splash:
        panel_wranger.use_splash_screen(splash)
        process_qt_events(app)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np

MAX_PRECISION = -24


def float_to_string(f, precision=2):
    """Create pretty floats for legend items"""
    fmt = f"{{:.{precision}f}}"
    return fmt.format(f).rstrip("0").rstrip(".")


def safe_log10(value):
    """Return the max precision if we receive zero when evaluating a value
       with log10. This is to protect the GUI from -inf."""

    if np.isscalar(value):
        return np.log10(value) if value > 0 else MAX_PRECISION

    if not isinstance(value, np.ndarray):
        value = np.array(value)

    # Calculate log10 for valid values, replace with nans otherwise
    index = value > 0
    result = np.log10(value, where=index)
    result[~index] = np.nan

    return result



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QLayout, QLayoutItem

from karabogui.sceneview.utils import save_painter_state


class _ShapeLayoutItemChild(QLayoutItem):
    """ A private QLayoutItem for the ShapeLayoutItem
    """
    def __init__(self, shape):
        super().__init__()
        self.shape = shape

    def isEmpty(self):
        """ This is part of the virtual interface of QLayoutItem.
        """
        return True


class ShapeLayoutItem(QLayout):
    """ A QLayout for holding a single shape object.

    This class exists because Qt apparently doesn't support QLayoutItem
    subclasses. However, since QLayout is a QLayoutItem, you can subclass it to
    get the same behavior when you want to add an item to a layout which isn't
    a layout or a widget (or a spacer). So here we are.
    """
    def __init__(self, shape):
        super().__init__()
        self._item = None
        self.addItem(_ShapeLayoutItemChild(shape))

    # --------------------------------------------
    # SceneView item interface

    @property
    def shape(self):
        return self._item.shape

    @property
    def model(self):
        return self._item.shape.model

    def draw(self, painter):
        shape = self._item.shape
        if shape.is_visible():
            with save_painter_state(painter):
                shape.draw(painter)

    def hide(self):
        self._item.shape.hide()

    def show(self):
        self._item.shape.show()

    def set_geometry(self, rect):
        self._item.shape.set_geometry(rect)

    def translate(self, offset):
        self._item.shape.translate(offset)

    # --------------------------------------------
    # QLayout Virtual Functions

    def addItem(self, item):
        self._item = item

    def count(self):
        return 0 if self._item is None else 1

    def geometry(self):
        return self._item.shape.geometry()

    def itemAt(self, index):
        if index == 0:
            return self._item

    def minimumSize(self):
        return self._item.shape.minimumSize()

    def maximumSize(self):
        return self._item.shape.maximumSize()

    def sizeHint(self):
        return self.minimumSize()

    def setGeometry(self, rect):
        self._item.shape.set_geometry(rect)

    def takeAt(self, index):
        if index == 0:
            self._item = None
        return None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.common.api import KARABO_SCHEMA_MAX_INC, KARABO_SCHEMA_MIN_INC
from karabogui import messagebox
from karabogui.binding.api import FloatBinding, StringBinding, Uint64Binding
from karabogui.controllers import icons_dialogs
from karabogui.testing import click_button


def test_digit_dialog_threshold(gui_app):
    MAX_UINT64 = 0xFFFFFFFFFFFFFFFF

    # 1. Floats
    binding = FloatBinding(attributes={
        KARABO_SCHEMA_MIN_INC: 3.0,
        KARABO_SCHEMA_MAX_INC: MAX_UINT64})
    dialog = icons_dialogs.DigitDialog([], binding)
    assert dialog.value.validator().minInc == 3.0
    assert dialog.value.validator().maxInc == MAX_UINT64

    # 2. Integers
    binding = Uint64Binding(attributes={
        KARABO_SCHEMA_MIN_INC: 2,
        KARABO_SCHEMA_MAX_INC: MAX_UINT64}
    )
    dialog = icons_dialogs.DigitDialog([], binding)
    assert dialog.value.validator().minInc == 2
    # Integers have a 32 bit limitation
    assert dialog.value.validator().maxInc == MAX_UINT64


def test_digit_dialog_order_of_items(gui_app, mocker):
    """Test the order of items added.They should always be sorted."""
    dialog = icons_dialogs.DigitDialog([], FloatBinding())
    dialog.lessEqual.setChecked(True)

    for val in (10.0, 0.0, 20.0):
        dialog.value.setText(str(val))
        click_button(dialog.addValue)
    valueList = dialog.valueList
    items = [valueList.item(i).text() for i in range(valueList.count())]
    assert items == ["<= 0.0", "<= 10.0", "<= 20.0"]

    # Add existing value with different operator - "<"
    dialog.less.setChecked(True)
    dialog.value.setText("10.0")
    click_button(dialog.addValue)
    items = [valueList.item(i).text() for i in range(valueList.count())]
    assert items == ["<= 0.0", "< 10.0", "<= 10.0", "<= 20.0"]

    # Add existing value with same operator.
    mock_error = mocker.patch.object(messagebox, "show_error")
    click_button(dialog.addValue)
    message = "Cannot add new condition; it already exists."
    mock_error.assert_called_once_with(message, parent=dialog)


def test_text_dialog_options(gui_app):
    """String property with options"""
    binding = StringBinding(
        attributes={"options": ["one", "two", "three"]})
    dialog = icons_dialogs.TextDialog(items=[], binding=binding)

    assert dialog.stack.currentWidget() == dialog.textOptionsPage
    value_list = dialog.valueList
    # The binding options are showed in a combobox and not added to the
    # list widget.
    assert value_list.count() == 0
    assert len(dialog.itemsComboBox) == 3
    for i in range(3):
        assert dialog.itemsComboBox.itemText(i) == binding.options[i]

    # The items in the controller's model are added to the list widget.
    item1 = icons_dialogs.IconItem(value="one")
    item2 = icons_dialogs.IconItem(value="two")

    dialog = icons_dialogs.TextDialog(items=[item1, item2],
                                      binding=binding)
    value_list = dialog.valueList
    assert value_list.count() == 2


def test_text_dialog_no_options(gui_app):
    """String property without options"""
    binding = StringBinding()
    dialog = icons_dialogs.TextDialog(items=[], binding=binding)
    assert dialog.stack.currentWidget() == dialog.textPage

    value_list = dialog.valueList
    items = [value_list.item(i).text() for i in range(value_list.count())]
    assert items == []



#############################################################################
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtWidgets import QVBoxLayout, QWidget

from karabogui.events import KaraboEvent, register_for_broadcasts
from karabogui.navigation.device_view import DeviceTreeView
from karabogui.singletons.api import get_config

from .base import BasePanelWidget
from .tool_widget import InterfaceBar


class DevicePanel(BasePanelWidget):
    def __init__(self):
        super().__init__("Device Topology")
        # We need broadcasts!
        event_map = {
            KaraboEvent.AccessLevelChanged: self._event_access_level,
            KaraboEvent.NetworkConnectStatus: self._event_network
        }
        register_for_broadcasts(event_map)

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        widget = QWidget(self)
        main_layout = QVBoxLayout(widget)
        main_layout.setContentsMargins(2, 2, 2, 2)
        self.tree_view = DeviceTreeView(widget)

        self.sbar = InterfaceBar(parent=widget)
        self.sbar.setView(self.tree_view)

        main_layout.addWidget(self.sbar)
        main_layout.addWidget(self.tree_view)

        return widget

    def _event_access_level(self, data):
        self.sbar.reset(enable=True)

    def _event_network(self, data):
        status = data.get('status', False)
        if not status:
            self.close_popup_widget()
        self.sbar.reset(status)

    def close_popup_widget(self):
        widget = self.tree_view.popupWidget
        if widget is not None:
            widget.close()

    def __repr__(self):
        topic = get_config()["broker_topic"]
        return f"<DevicePanel topic={topic}>"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui import icons

from ..enums import ExportTool, MouseTool, ROITool
from .widgets import create_action, create_button, create_dropdown_button


def tool_factory(tool):
    factory = FACTORY_MAP.get(tool.__class__)
    if factory is None:
        return

    return factory(tool)


def mouse_mode_factory(tool):
    """Add optional buttons"""
    button = None
    if tool is MouseTool.Pointer:
        button = create_button(icon=icons.pointer,
                               checkable=True,
                               tooltip="Pointer Mode")
    elif tool is MouseTool.Zoom:
        button = create_button(icon=icons.zoomImage,
                               checkable=True,
                               tooltip="Zoom Mode")
    elif tool is MouseTool.Move:
        button = create_button(icon=icons.move,
                               checkable=True,
                               tooltip="Move Mode")
    elif tool is MouseTool.Picker:
        button = create_button(icon=icons.picker,
                               checkable=True,
                               tooltip="Picker Mode")
    return button


def roi_factory(tool):
    button = None
    if tool is ROITool.Rect:
        # Create Rect ROI dropdown button
        show_rect_action = create_action(text="Show Rect ROI",
                                         icon=icons.roi,
                                         data=ROITool.Rect,
                                         checkable=True)
        draw_rect_action = create_action(text="Draw Rect ROI",
                                         icon=icons.drawROI,
                                         data=ROITool.DrawRect,
                                         checkable=True)

        button = create_dropdown_button(
            actions=[show_rect_action, draw_rect_action],
            icon=icons.roi,
            tooltip="Rect ROI",
            checkable=True)
    elif tool is ROITool.Crosshair:
        # Create Crosshair ROI dropdown button
        show_crosshair_action = create_action(text="Show Crosshair ROI",
                                              icon=icons.crosshair,
                                              data=ROITool.Crosshair,
                                              checkable=True)
        draw_crosshair_action = create_action(text="Draw Crosshair ROI",
                                              icon=icons.drawCrosshair,
                                              data=ROITool.DrawCrosshair,
                                              checkable=True)
        button = create_dropdown_button(
            actions=[show_crosshair_action, draw_crosshair_action],
            icon=icons.crosshair,
            tooltip="Rect ROI",
            checkable=True)
    return button


def export_factory(tool):
    button = None
    if tool is ExportTool.Image:
        button = create_button(icon=icons.camera,
                               checkable=False,
                               tooltip="Save Snapshot")
    elif tool is ExportTool.Data:
        button = create_button(icon=icons.export,
                               checkable=False,
                               tooltip="Export Data")
    return button


FACTORY_MAP = {
    ExportTool: export_factory,
    MouseTool: mouse_mode_factory,
    ROITool: roi_factory
}



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on March 6, 2021
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import webbrowser

from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import (
    QApplication, QComboBox, QMessageBox, QStyle, QStyledItemDelegate,
    QStyleOptionFrame, QStyleOptionProgressBar)

from karabo.common.api import KARABO_SCHEMA_OPTIONS
from karabo.native import Hash, is_equal
from karabogui import messagebox
from karabogui.binding.api import (
    BoolBinding, FloatBinding, IntBinding, StringBinding, VectorBinding,
    VectorBoolBinding, get_default_value, get_min_max)
from karabogui.controllers.validators import (
    BindingValidator as GenericValidator, ListValidator, SimpleValidator)
from karabogui.logger import get_logger
from karabogui.request import (
    call_device_slot, get_scene_from_server, retrieve_default_scene)
from karabogui.topology.api import is_device_online
from karabogui.util import SignalBlocker, get_reason_parts
from karabogui.widgets.edits import LineEditEditor

from .button_delegate import TableButtonDelegate
from .utils import (
    create_brushes, get_button_attributes, has_confirmation, parse_table_link)
from .widgets import TableVectorEdit


def get_display_delegate(proxy, binding, parent, read_only=True):
    """Retrieve a display (readonly) delegate for the column with binding

    :param proxy: The property proxy of the table element
    :param binding: The column binding
    :param parent: The parent widget
    :param read_only: If only read_only display delegates are considered
    """
    displayType = binding.displayType.split("|")[0]
    if not displayType:
        return None
    if (displayType == "TableBoolButton" and read_only and
            isinstance(binding, BoolBinding)):
        return BoolButtonDelegate(proxy, binding, parent)
    elif (displayType == "TableStringButton" and read_only and
          isinstance(binding, StringBinding)):
        return StringButtonDelegate(proxy, binding, parent)
    elif (displayType == "TableProgressBar"
          and isinstance(binding, (FloatBinding, IntBinding))):
        return ProgressBarDelegate(binding, parent)
    elif displayType == "TableColor":
        if isinstance(binding, (FloatBinding, IntBinding)):
            return ColorNumberDelegate(binding, parent)
        elif isinstance(binding, StringBinding):
            return ColorBindingDelegate(binding, parent)


def get_table_delegate(proxy, binding, parent):
    """Retrieve a (writable) table delegate for the column with binding

    :param proxy: The property proxy of the table element
    :param binding: The column binding
    :param parent: The parent widget
    """
    # For a writable table we might find other delegates allowed
    delegate = get_display_delegate(proxy, binding, parent,
                                    read_only=False)
    if delegate is not None:
        return delegate

    options = binding.attributes.get(KARABO_SCHEMA_OPTIONS, None)
    if options is not None:
        return ComboBoxDelegate(options, parent)
    elif isinstance(binding, (FloatBinding, IntBinding)):
        return NumberDelegate(binding, parent)
    elif isinstance(binding, VectorBinding):
        if (binding.displayType.split("|")[0] == "TableVector"
                and not isinstance(binding, VectorBoolBinding)):
            return VectorButtonDelegate(binding, parent)
        return VectorDelegate(binding, parent)
    else:
        return BindingDelegate(binding, parent)


class VectorButtonDelegate(QStyledItemDelegate):

    def __init__(self, binding, parent=None):
        super().__init__(parent)
        self.binding = binding

    def createEditor(self, parent, option, index):
        editor = TableVectorEdit(self.binding, parent=parent)
        values = index.model().data(index, Qt.DisplayRole)
        old = _create_string_list(values)
        validator = VectorValidator(self.binding, old, parent=editor)
        editor.setValidator(validator)
        return editor

    def setEditorData(self, editor, index):
        """Reimplemented function of QStyledItemDelegate"""
        values = index.model().data(index, Qt.DisplayRole)
        editor.setText(values)

    def setModelData(self, editor, model, index):
        """Reimplemented function of QStyledItemDelegate"""
        old = index.model().data(index, Qt.DisplayRole)
        new = editor.text()
        if not is_equal(old, new):
            model.setData(index, new, Qt.EditRole)


class BoolButtonDelegate(TableButtonDelegate):
    def __init__(self, proxy, binding, parent=None):
        super().__init__(parent)
        self.text = binding.displayedName or proxy.path
        self.deviceId = proxy.root_proxy.device_id
        self.path = proxy.path
        self.confirm = has_confirmation(binding.displayType)

    def isEnabled(self, index=None):
        """Reimplemented function of TableButtonDelegate"""
        enabled = index.data(role=Qt.CheckStateRole) == Qt.Checked
        return enabled

    def get_button_text(self, index):
        """Reimplemented function of TableButtonDelegate"""
        return self.text

    def click_action(self, index):
        """Reimplemented function of TableButtonDelegate"""
        if not index.isValid() or not is_device_online(self.deviceId):
            return

        if self.confirm:
            message_box = QMessageBox()
            message_box.setModal(False)
            confirm = message_box.question(
                self.parent(), "Confirmation",
                f"Do you really want to proceed with <b>{self.text}</b>?",
                QMessageBox.No, QMessageBox.Yes)
            if confirm == QMessageBox.No:
                return

        model = index.model()
        row = index.row()
        column = index.column()
        header = model.get_header_key(column)

        column_count = model.columnCount()
        h = Hash(model.get_model_data(row, col) for col in range(column_count))
        data = Hash("rowData", h, "row", row, "column", column,
                    "header", header)

        def request_handler(success, reply):
            if not success:
                reason, details = get_reason_parts(reply)
                messagebox.show_error("Request for table action for device "
                                      f"<b>{self.deviceId}</b> failed."
                                      "The reason is:<br>"
                                      f"<i>{reason}</i>", details=details,
                                      parent=self.parent())
            elif not reply.get("payload.success", True):
                reason = reply.get("payload.reason", "")
                messagebox.show_error(
                    "Request for table action for device "
                    f"<b>{self.deviceId}</b> reported an error "
                    f"with the reason: {reason}.", parent=self.parent())
            else:
                get_logger().info("Successfully executed action "
                                  f"<b>{self.text}</b> for device "
                                  f"<b>{self.deviceId}</b>.")

        call_device_slot(request_handler, self.deviceId,
                         "requestAction", action="TableButton",
                         path=self.path, table=data)


class StringButtonDelegate(TableButtonDelegate):
    def __init__(self, proxy, binding, parent=None):
        super().__init__(parent)
        self.text = binding.displayedName or proxy.path
        self.deviceId = proxy.root_proxy.device_id
        self.attributes = get_button_attributes(binding.displayType)

    def get_button_text(self, index):
        """Reimplemented function of TableButtonDelegate"""
        return self.text

    def click_action(self, index):
        """Reimplemented function of TableButtonDelegate"""
        if not index.isValid() or not is_device_online(self.deviceId):
            return

        model = index.model()
        _, data = model.get_model_data(index.row(), index.column())

        action, options = parse_table_link(data)
        if action is None:
            get_logger().debug(
                "Invalid scheme provided for table string button")
        elif action == "deviceScene":
            try:
                deviceId = options["device_id"]
                name = options.get("name")
                if name is None:
                    retrieve_default_scene(deviceId)
                else:
                    get_scene_from_server(deviceId, name)
            except Exception:
                get_logger().warning(
                    f"Invalid deviceScene information: {data}.")
        elif action == "url":
            try:
                webbrowser.open_new(options)
            except Exception:
                get_logger().error(f"No valid url specified {options}")


class ComboBoxDelegate(QStyledItemDelegate):
    def __init__(self, options, parent=None):
        super().__init__(parent)
        # Note: We make sure that the options are a list of strings. For simple
        # types in Karabo they are a coercing array on the binding
        self._options = [str(value) for value in options]

    def createEditor(self, parent, option, index):
        """Reimplemented function of QStyledItemDelegate"""
        editor = QComboBox(parent)
        editor.addItems(self._options)
        editor.currentIndexChanged.connect(self._on_editor_changed)
        return editor

    def setEditorData(self, editor, index):
        """Reimplemented function of QStyledItemDelegate"""
        selection = index.model().data(index, Qt.DisplayRole)
        if selection in self._options:
            selection_index = self._options.index(selection)
            with SignalBlocker(editor):
                editor.setCurrentIndex(selection_index)
        else:
            with SignalBlocker(editor):
                editor.setCurrentIndex(-1)
            get_logger().error(
                f"The value {selection} is not in the following "
                f"options: {self._options}")

    def setModelData(self, editor, model, index):
        """Reimplemented function of QStyledItemDelegate"""
        old = index.model().data(index, Qt.DisplayRole)
        new = self._options[editor.currentIndex()]
        if not is_equal(old, new):
            model.setData(index, new, Qt.EditRole)

    @Slot()
    def _on_editor_changed(self):
        """The current index of the combobox changed. Notify the model.

        This signal MUST be emitted when the editor widget has completed
        editing the data, and wants to write it back into the model.

        XXX: This is in principle a wrong implementation, as it should be
        only emitted when the editor finished.
        """
        self.commitData.emit(self.sender())


class NumberDelegate(QStyledItemDelegate):
    def __init__(self, binding, parent=None):
        super().__init__(parent)
        self._binding = binding

    def createEditor(self, parent, option, index):
        """Reimplemented function of QStyledItemDelegate"""
        editor = LineEditEditor(parent)
        old = index.model().data(index, Qt.DisplayRole)
        validator = NumberValidator(self._binding, old, parent=editor)
        editor.setValidator(validator)
        return editor

    def setModelData(self, editor, model, index):
        """Reimplemented function of QStyledItemDelegate

        We only recreate the table if we have changes in our value! We can do
        this since the delegate will call `setModelData` once its out of focus.
        """
        old = index.model().data(index, Qt.DisplayRole)
        new = editor.text()
        if not is_equal(old, new):
            model.setData(index, new, Qt.EditRole)
            self.commitData.emit(self.sender())


class ProgressBarDelegate(NumberDelegate):

    def __init__(self, binding, parent=None):
        super().__init__(binding, parent)
        self.min_value, self.max_value = get_min_max(binding)
        self.has_limits = (self.min_value is not None
                           and self.max_value is not None)
        self.editing = {}

    def createEditor(self, parent, option, index):
        """Reimplemented function of QStyledItemDelegate."""
        key = (index.row(), index.column())
        self.editing[key] = True
        return super().createEditor(parent, option, index)

    def destroyEditor(self, editor, index):
        """Reimplemented function of QStyledItemDelegate."""
        key = (index.row(), index.column())
        self.editing.pop(key, None)
        super().destroyEditor(editor, index)

    def paint(self, painter, option, index):
        """Reimplemented function of QStyledItemDelegate."""
        key = (index.row(), index.column())
        if self.editing.get(key):
            super().paint(painter, option, index)
            return

        progress_bar = QStyleOptionProgressBar()
        progress_bar.rect = option.rect
        if self.has_limits:
            progress_bar.minimum = int(self.min_value)
            progress_bar.maximum = int(self.max_value)
            progress_bar.textVisible = True
            progress_bar.text = index.data(role=Qt.DisplayRole)
            progress_bar.progress = int(float(index.data(role=Qt.EditRole)))
        QApplication.style().drawControl(QStyle.CE_ProgressBar, progress_bar,
                                         painter)


class NumberValidator(SimpleValidator):

    def __init__(self, binding, old, parent=None):
        super().__init__(binding=binding, parent=parent)
        self._old_value = old

    def fixup(self, input):
        """Reimplemented function of QValidator

        If the value input has not been validated properly return the previous
        value before editing.
        """
        if self._old_value is None or self._old_value == "":
            # Note: Ideally a table value should always be there. We make sure
            # it stays like this
            return str(get_default_value(self._binding, force=True))

        return str(self._old_value)


class VectorDelegate(QStyledItemDelegate):
    def __init__(self, binding, parent=None):
        super().__init__(parent=parent)
        self._binding = binding

    def createEditor(self, parent, option, index):
        """Reimplemented function of QStyledItemDelegate"""
        editor = LineEditEditor(parent)
        old = index.model().data(index, Qt.DisplayRole)
        old = _create_string_list(old)
        validator = VectorValidator(self._binding, old, parent=editor)
        editor.setValidator(validator)
        return editor

    def setModelData(self, editor, model, index):
        """Reimplemented function of QStyledItemDelegate"""
        old = index.model().data(index, Qt.DisplayRole)
        old = _create_string_list(old)
        new = editor.text()
        if not is_equal(old, new):
            model.setData(index, new, Qt.EditRole)
            self.commitData.emit(self.sender())

    def setEditorData(self, editor, index):
        """Reimplemented function of QStyledItemDelegate"""
        value = index.model().data(index, Qt.DisplayRole)
        value = _create_string_list(value)
        editor.setText(value)


def _create_string_list(value):
    """Return a string list with stripped white spaces if present"""
    return value if not value else ",".join(
        [v.rstrip().lstrip() for v in value.split(",")])


class VectorValidator(ListValidator):

    def __init__(self, binding, old, parent=None):
        super().__init__(binding=binding, parent=parent)
        self._old_value = old

    def fixup(self, input):
        """Reimplemented function of QValidator"""
        if self._old_value is None:
            # The model will account for an empty string and use a list
            return ""

        return str(self._old_value)


class BindingDelegate(QStyledItemDelegate):
    def __init__(self, binding, parent=None):
        super().__init__(parent)
        self._binding = binding

    def createEditor(self, parent, option, index):
        """Reimplemented function of QStyledItemDelegate"""
        editor = LineEditEditor(parent)
        old = index.model().data(index, Qt.DisplayRole)
        validator = BindingValidator(self._binding, old, parent=editor)
        editor.setValidator(validator)
        return editor

    def setModelData(self, editor, model, index):
        """Reimplemented function of QStyledItemDelegate"""
        old = index.model().data(index, Qt.DisplayRole)
        new = editor.text()
        if not is_equal(old, new):
            model.setData(index, new, Qt.EditRole)
            self.commitData.emit(self.sender())


class BindingValidator(GenericValidator):

    def __init__(self, binding, old, parent=None):
        super().__init__(binding=binding, parent=parent)
        self._old_value = old

    def fixup(self, input):
        """Reimplemented function of QValidator"""
        if self._old_value is None:
            return str(get_default_value(self._binding, force=True))

        return str(self._old_value)


def _color_delegate(name, base):
    """Create a color delegate with a different base class"""

    class ColorDelegate:

        def __init__(self, binding, parent=None):
            super().__init__(binding, parent)
            self.editing = {}
            self.default_brush, self.brushes = create_brushes(
                binding.displayType)

        def createEditor(self, parent, option, index):
            """Reimplemented function of QStyledItemDelegate."""
            key = (index.row(), index.column())
            self.editing[key] = True
            return super().createEditor(parent, option, index)

        def destroyEditor(self, editor, index):
            """Reimplemented function of QStyledItemDelegate."""
            key = (index.row(), index.column())
            self.editing.pop(key, None)
            super().destroyEditor(editor, index)

        def paint(self, painter, option, index):
            """Reimplemented function of QStyledItemDelegate."""
            key = (index.row(), index.column())
            if self.editing.get(key):
                super().paint(painter, option, index)
                return

            frame = QStyleOptionFrame()
            rect = option.rect
            frame.rect = rect
            text = index.data(role=Qt.DisplayRole)
            brush = self.brushes.get(text, self.default_brush)
            if brush is not None:
                painter.fillRect(rect, brush)
            painter.drawText(rect, Qt.AlignCenter, text)
            QApplication.style().drawControl(QStyle.CE_ShapedFrame, frame,
                                             painter)

    return type(name, (ColorDelegate, base), {})


ColorBindingDelegate = _color_delegate("ColorBindingDelegate", BindingDelegate)
ColorNumberDelegate = _color_delegate("ColorNumberDelegate", NumberDelegate)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QModelIndex, Qt
from qtpy.QtWidgets import QDialog, QMessageBox

from karabo.native import AccessLevel, Timestamp
from karabogui.dialogs.api import (
    InitConfigurationDialog, SaveConfigurationDialog)
from karabogui.events import KaraboEvent
from karabogui.singletons.mediator import Mediator
from karabogui.testing import (
    access_level, click_button, singletons, system_hash)
from karabogui.topology.api import SystemTopology


def get_config_items():
    items = []
    for i in range(4):
        config = {}
        config["serverId"] = "swerver"
        config["classId"] = "FooClass"
        config["name"] = f"default{i}"
        config["date"] = Timestamp().toLocal(" ", "seconds")
        items.append(config)
    return items


def test_init_config_dialog(gui_app, mocker):
    mediator = Mediator()
    network = mocker.Mock()
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(mediator=mediator, network=network, topology=topology):
        dialog = InitConfigurationDialog(instance_id="divvy")
        assert dialog.ui_instance.text() == "Device Id: divvy"
        # List of configurations is asked for!
        assert network.onListInitConfigurations.call_count == 1

        # Clicking the show device requests navigation panel to show
        path = "karabogui.dialogs.init_configuration.broadcast_event"
        broadcast = mocker.patch(path)
        click_button(dialog.ui_show_device)
        broadcast.assert_called_with(
            KaraboEvent.ShowDevice, {"deviceId": "divvy",
                                     "showTopology": True})

        # Let configurations arrive! First empty!
        event = {"items": [], "deviceId": "divvy"}
        dialog._event_list_updated(event)
        assert dialog.ui_status.text() == "No configurations are available!"

        # Now a set of configurations
        event = {"items": get_config_items(), "deviceId": "divvy"}
        dialog._event_list_updated(event)
        assert dialog.ui_status.text() == ""

        model = dialog.model
        source_model = model.sourceModel()
        assert model.rowCount(QModelIndex()) == 4
        assert source_model.rowCount(QModelIndex()) == 4

        # Change a filter setting
        dialog.ui_filter.setText("1")
        assert model.rowCount(QModelIndex()) == 1
        assert source_model.rowCount(QModelIndex()) == 4

        # Select a cell and request
        dialog.ui_table_widget.selectRow(0)
        # XXX: Unfortunately, QTest doubleClick is not working [Bug]
        dialog._request_configuration()
        network.onGetInitConfiguration.assert_called_with(
            "divvy", "default1", False)

        # Click refresh button
        click_button(dialog.ui_button_refresh)
        network.onListInitConfigurations.assert_called_with("divvy")

        # ContextMenu
        click_button(dialog.ui_table_widget, Qt.RightButton)

        # Delete action
        QMessageBox.exec = mocker.Mock(return_value=QMessageBox.No)
        dialog._request_delete()
        network.onDeleteInitConfiguration.call_count == 0

        QMessageBox.exec = mocker.Mock(return_value=QMessageBox.Yes)
        dialog._request_delete()
        network.onDeleteInitConfiguration.assert_called_with(
            "divvy", "default1")

        # Unregister and set successful!
        dialog.done(1)

        # Network event and close
        dialog._event_network({"status": False})

        index = source_model.index(0, 0)
        assert index.data() == "default0"

        save = "karabogui.dialogs.init_configuration." \
               "SaveConfigurationDialog"
        dia = mocker.patch(save)
        dia().name = "Next"
        dia().exec.return_value = QDialog.Accepted
        # Call it!
        dialog.open_save_dialog()
        network.onSaveInitConfiguration.assert_called_with(
            "Next", ["divvy"], update=True)


def test_config_save_dialog(gui_app, mocker):
    """Test the save config dialog"""
    mediator = Mediator()
    with singletons(mediator=mediator):
        dialog = SaveConfigurationDialog()

        # Network event and close
        mediator.unregister_listener = mocker.Mock()
        dialog.done(1)
        mediator.unregister_listener.assert_called_once()
        # Asserts simple defaults
        assert dialog.name == ""

        # Network event should be there and closes
        dialog._event_network({"status": False})


def test_delete_configuration(gui_app, mocker):
    network = mocker.Mock()
    with singletons(network=network), access_level(AccessLevel.EXPERT):
        dialog = InitConfigurationDialog(instance_id="divvy")
        assert not dialog.ui_button_delete.isEnabled()
        event = {"items": get_config_items(), "deviceId": "divvy"}
        dialog._event_list_updated(event)

        dialog.ui_table_widget.selectRow(0)
        assert dialog.ui_button_delete.isEnabled()
        QMessageBox.exec = mocker.Mock(return_value=QMessageBox.Yes)
        click_button(dialog.ui_button_delete)
        network.onDeleteInitConfiguration.assert_called_with(
            "divvy", "default0")



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 27, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from functools import partial
from io import StringIO

from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QAction, QDialog, QMenu, QMessageBox
from traits.api import Instance, Property, on_trait_change

import karabogui.icons as icons
from karabo.common.api import Capabilities
from karabo.common.project.api import (
    DeviceConfigurationModel, DeviceInstanceModel, DeviceServerModel,
    find_parent_object)
from karabo.native import read_project_model, write_project_model
from karabogui import messagebox
from karabogui.access import (
    AccessRole, access_role_allowed, get_access_level_for_role)
from karabogui.binding.api import (
    NO_CLASS_STATUSES, NO_CONFIG_STATUSES, ONLINE_CONFIG_STATUSES)
from karabogui.const import TOOLTIP_INIT_CONFIG, TOOLTIP_RUNTIME_CONFIG
from karabogui.dialogs.api import (
    ConfigComparisonDialog, ConfigurationFromPastDialog,
    DeviceCapabilityDialog, InitConfigurationDialog)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.indicators import get_project_device_status_icon
from karabogui.itemtypes import ProjectItemTypes
from karabogui.logger import get_logger
from karabogui.project.dialog.device_configuration import (
    DeviceConfigurationDialog)
from karabogui.project.dialog.device_handle import DeviceHandleDialog
from karabogui.project.dialog.object_handle import ObjectDuplicateDialog
from karabogui.project.utils import check_device_instance_exists
from karabogui.request import (
    get_macro_from_server, get_scene_from_server, retrieve_default_scene)
from karabogui.singletons.api import get_config, get_manager, get_topology
from karabogui.topology.api import ProjectDeviceInstance
from karabogui.util import move_to_cursor

from .bases import BaseProjectController, ProjectControllerUiData
from .server import DeviceServerController

ACCESS_LEVEL_TOOLTIP = "Requires minimum '{}' access level"
SERVER_OFFLINE_TOOLTIP = "Server is offline"
DEVICE_ONLINE_TOOLTIP = "Device is already instantiated"
DEVICE_OFFLINE_TOOLTIP = "Device is offline"
INCORRECT_STATE_TOOLTIP = "Device is not in correct State"


class DeviceInstanceController(BaseProjectController):
    """ A wrapper for DeviceInstanceModel objects
    """
    # Redefine model with the correct type
    model = Instance(DeviceInstanceModel)
    active_config = Property(Instance(DeviceConfigurationModel))

    # A proxy for online and offline device configurations
    project_device = Instance(ProjectDeviceInstance)

    def context_menu(self, project_controller, parent=None):
        def _test_mask(mask, bit):
            return (mask & bit) == bit

        menu = QMenu(parent)
        menu.setToolTipsVisible(True)
        server_controller = find_parent_object(self, project_controller,
                                               DeviceServerController)
        server_online = server_controller.online
        proj_device_online = self.project_device.online
        proj_device_status = self.project_device.proxy.status
        proj_topo_node = self.project_device.device_node

        capabilities = proj_topo_node.capabilities if proj_topo_node else 0

        project_allowed = access_role_allowed(AccessRole.PROJECT_EDIT)
        service_allowed = access_role_allowed(AccessRole.SERVICE_EDIT)

        project_edit_access_level = get_access_level_for_role(
            AccessRole.PROJECT_EDIT)
        project_allowed_tooltip = ACCESS_LEVEL_TOOLTIP.format(
            project_edit_access_level)

        service_edit_access_level = get_access_level_for_role(
            AccessRole.SERVICE_EDIT)
        service_allowed_tooltip = ACCESS_LEVEL_TOOLTIP.format(
            service_edit_access_level)

        edit_action = QAction(icons.edit, 'Edit', menu)
        edit_action.triggered.connect(partial(self._edit_device,
                                              project_controller,
                                              parent=parent))
        edit_action.setEnabled(project_allowed)
        if not project_allowed:
            edit_action.setToolTip(project_allowed_tooltip)

        config_menu = self._create_sub_menu(menu, project_controller,
                                            project_allowed)
        dupe_action = QAction(icons.editCopy, 'Duplicate', menu)
        dupe_action.triggered.connect(partial(self._duplicate_device,
                                              project_controller,
                                              parent=parent))
        dupe_action.setEnabled(project_allowed)
        if not project_allowed:
            dupe_action.setToolTip(project_allowed_tooltip)

        delete_action = QAction(icons.delete, 'Delete', menu)
        delete_action.triggered.connect(partial(self._delete_device,
                                                project_controller,
                                                parent=parent))
        delete_action.setEnabled(project_allowed)
        if not project_allowed:
            delete_action.setToolTip(project_allowed_tooltip)
        macro_action = QAction(icons.download, 'Open device macro', menu)
        has_macro = _test_mask(capabilities, Capabilities.PROVIDES_MACROS)
        macro_action.triggered.connect(partial(self._load_macro_from_device,
                                               project_controller,
                                               parent=parent))
        macro_action.setEnabled(has_macro and project_allowed)
        if not has_macro:
            macro_action.setToolTip("The device has no Macro")
        elif not project_allowed:
            macro_action.setToolTip(project_allowed_tooltip)

        scene_action = QAction(icons.download, 'Open device scene', menu)
        has_scene = _test_mask(capabilities, Capabilities.PROVIDES_SCENES)
        scene_action.triggered.connect(partial(self._load_scene_from_device,
                                               project_controller,
                                               parent=parent))
        scene_action.setEnabled(has_scene and project_allowed)
        if not has_scene:
            scene_action.setToolTip("Device has no scene")
        elif not project_allowed:
            scene_action.setToolTip(project_allowed_tooltip)

        conf_action = QAction(icons.clock, 'Runtime configuration', menu)
        can_get_conf = (server_online and
                        proj_device_status not in NO_CONFIG_STATUSES)
        conf_action.triggered.connect(partial(
            self._get_configuration_from_past, parent=parent))
        conf_action.setToolTip(TOOLTIP_RUNTIME_CONFIG)
        conf_action.setEnabled(can_get_conf)
        if not server_online:
            conf_action.setToolTip(SERVER_OFFLINE_TOOLTIP)
        elif not can_get_conf:
            conf_action.setToolTip(INCORRECT_STATE_TOOLTIP)

        conf_action_name = QAction(icons.initconfig, 'Init configuration',
                                   menu)
        can_get_conf_name = (server_online and
                             proj_device_status not in NO_CONFIG_STATUSES)
        conf_action_name.triggered.connect(partial(
            self._get_configuration_from_name, parent=parent))
        conf_action_name.setEnabled(can_get_conf_name)
        conf_action_name.setToolTip(TOOLTIP_INIT_CONFIG)
        if not server_online:
            conf_action.setToolTip(SERVER_OFFLINE_TOOLTIP)
        elif not can_get_conf:
            conf_action.setToolTip(INCORRECT_STATE_TOOLTIP)

        show_action = QAction(icons.deviceInstance, 'Select in topology', menu)
        show_action.triggered.connect(self._show_device)
        show_action.setVisible(proj_device_online)

        set_conf_action = QAction('Store online configuration', menu)
        set_can_get_conf = (server_online and
                            proj_device_status in ONLINE_CONFIG_STATUSES and
                            project_allowed)
        set_conf_action.setEnabled(set_can_get_conf)
        if not server_online:
            set_conf_action.setToolTip(SERVER_OFFLINE_TOOLTIP)
        elif not project_allowed:
            set_conf_action.setToolTip(project_allowed_tooltip)
        elif not set_can_get_conf:
            set_conf_action.setToolTip(INCORRECT_STATE_TOOLTIP)
        set_conf_action.triggered.connect(partial(
            self._store_online_configuration, parent=parent))

        instantiate_action = QAction(icons.run, 'Instantiate', menu)

        can_instantiate = (server_online and not proj_device_online and
                           proj_device_status not in NO_CLASS_STATUSES)
        instantiate_action.triggered.connect(partial(self._instantiate_device,
                                                     project_controller,
                                                     parent=parent))
        instantiate_action.setEnabled(can_instantiate and service_allowed)
        if not server_online:
            instantiate_action.setToolTip(INCORRECT_STATE_TOOLTIP)
        elif proj_device_online:
            instantiate_action.setToolTip(DEVICE_ONLINE_TOOLTIP)
        elif not can_instantiate:
            instantiate_action.setToolTip(INCORRECT_STATE_TOOLTIP)
        elif not service_allowed:
            instantiate_action.setToolTip(service_allowed_tooltip)

        shutdown_action = QAction(icons.kill, 'Shutdown', menu)
        shutdown_action.triggered.connect(partial(self.shutdown_device,
                                                  show_confirm=True,
                                                  parent=parent))
        shutdown_action.setEnabled(proj_device_online and service_allowed)
        if not proj_device_online:
            shutdown_action.setToolTip(DEVICE_OFFLINE_TOOLTIP)
        elif not service_allowed:
            shutdown_action.setToolTip(service_allowed_tooltip)

        menu.addSeparator()
        about_action = QAction(icons.about, 'About', menu)
        about_action.triggered.connect(partial(self._about_device,
                                               parent=parent))

        up_action = QAction(icons.arrowFancyUp, 'Move up', menu)
        up_action.triggered.connect(partial(self._move_up,
                                            project_controller))
        up_action.setEnabled(project_allowed)
        if not project_allowed:
            up_action.setToolTip(project_allowed_tooltip)

        down_action = QAction(icons.arrowFancyDown, 'Move down', menu)
        down_action.triggered.connect(partial(self._move_down,
                                              project_controller))
        down_action.setEnabled(project_allowed)
        if not project_allowed:
            down_action.setToolTip(project_allowed_tooltip)

        menu.addAction(edit_action)
        menu.addMenu(config_menu)
        menu.addSeparator()
        menu.addAction(dupe_action)
        menu.addAction(delete_action)
        menu.addSeparator()
        menu.addAction(macro_action)
        menu.addAction(scene_action)
        menu.addAction(conf_action)
        menu.addAction(conf_action_name)
        menu.addAction(show_action)
        menu.addAction(set_conf_action)
        menu.addSeparator()
        menu.addAction(instantiate_action)
        menu.addAction(shutdown_action)
        menu.addSeparator()
        menu.addAction(about_action)
        menu.addAction(up_action)
        menu.addAction(down_action)

        return menu

    def info(self):
        topology_node = self.project_device.device_node
        capabilities = topology_node.capabilities if topology_node else 0

        return {'type': ProjectItemTypes.DEVICE,
                'classId': self.model.class_id,
                'deviceId': self.model.instance_id,
                'capabilities': capabilities}

    def create_ui_data(self):
        ui_data = ProjectControllerUiData()
        self._update_icon(ui_data)
        self._update_instance_status(ui_data)
        return ui_data

    def single_click(self, project_controller, parent=None):
        if not self.model.initialized:
            return

        active_config = self.active_config
        if active_config is None:
            return

        self._broadcast_item_click()

    def double_click(self, project_controller, parent=None):
        project_device = self.project_device
        if project_device is None:
            return
        retrieve_default_scene(project_device.device_id)

    def delete_press(self, project_controller, parent=None):
        """Reimplemented function on `BaseProjectController`"""
        self._delete_device(project_controller, parent)

    def _get_display_name(self):
        """Traits property getter for ``display_name``
        """
        return self.model.instance_id

    # ----------------------------------------------------------------------
    # Traits handlers

    def _get_active_config(self):
        """Traits Property getter for the active device configuration object

        :return If a `DeviceConfigurationModel` for the `DeviceInstanceModel`s
        active `uuid` can be found, otherwise a `NoneType` is returned
        """
        device = self.model
        return device.select_config(device.active_config_ref)

    def _project_device_default(self):
        """Traits default initializer for `project_device`.
        """
        device = self.model
        config = self.active_config

        # Complain loudly if this trait is initialized when the model isn't
        assert device.initialized, "DeviceInstanceModel must be initialized!"
        assert config.initialized, "Device config must be initialized!"

        return get_topology().get_project_device(
            device.instance_id, device.server_id, device.class_id,
            None if config is None else config.configuration)

    @on_trait_change("model:instance_id,model:server_id,model:class_id")
    def _forced_project_device_change(self):
        """When instance_id/server_id/class_id changes, the project device
        needs to be notified
        """
        # Watch for incomplete model initialization
        if not self.model.initialized:
            return

        # Tell the project device
        self.project_device.rename(
            device_id=self.model.instance_id,
            server_id=self.model.server_id,
            class_id=self.model.class_id)

    @on_trait_change("model:initialized")
    def _update_ui_label(self):
        """ Whenever the model is initialized it should be visible to the user
        """
        self._update_icon(self.ui_data)
        self._update_instance_status(self.ui_data)

    @on_trait_change("model:active_config_ref")
    def _active_config_ref_change(self):
        """Called when user selects a different configuration in project panel
        """
        # Watch for incomplete model and view initialization
        if not self.model.initialized:
            return
        config = self.active_config
        if config is not None:
            self.project_device.set_project_config_hash(config.configuration)

    @on_trait_change('project_device:save_project')
    def _active_config_changed_in_configurator(self):
        """Called whenever project device's offline configuration is changed
        """
        if self.project_device.online:
            return
        config = self.active_config
        if config is not None:
            new_config = self.project_device.get_user_edited_config_hash()
            config.configuration = new_config

    @on_trait_change("project_device:status")
    def status_change(self, status):
        self._update_icon(self.ui_data)
        # Show the device's configuration, if it was already showing
        self._update_configurator()

    @on_trait_change("project_device:instance_status")
    def instance_status_change(self, status):
        self._update_instance_status(self.ui_data)

    # ----------------------------------------------------------------------
    # Util methods

    def active_config_changed(self, config_model):
        """Whenever the active config is changed from the context menu, update
        the project device proxy
        """
        if not self.model.initialized or config_model.configuration is None:
            return

        device = self.model
        # changing model's active_config_ref will trigger trait change handler
        device.active_config_ref = config_model.uuid
        # Notify configurator to display the new configuration,
        # ProjectDeviceInstance doesn't broadcast anything to configurator
        self._broadcast_item_click()

    def _broadcast_item_click(self):
        """Broadcast the request to view the project device

        This is not a lazy action. The device can be either online or offline
        and must be shown with a schema.
        """
        proxy = self.project_device.proxy
        if not self.project_device.online:
            proxy.ensure_class_schema()

        broadcast_event(KaraboEvent.ShowConfiguration, {'proxy': proxy})

    def _update_icon(self, ui_data):
        # Get current status of device
        if not self.model.initialized:
            return

        status_enum = self.project_device.status
        ui_data.icon = get_project_device_status_icon(status_enum)
        ui_data.status = status_enum

    def _update_instance_status(self, ui_data):
        if not self.model.initialized:
            return
        ui_data.instance_status = self.project_device.instance_status

    def _update_configurator(self, refresh=False):
        broadcast_event(KaraboEvent.UpdateDeviceConfigurator,
                        {'proxy': self.project_device.proxy,
                         "refresh": refresh})

    def _create_sub_menu(self, parent_menu, project_controller, allowed):
        """ Create sub menu for parent menu and return it
        """
        config_menu = QMenu('Configuration', parent_menu)
        show_action = QAction('Show configuration', config_menu)
        show_action.triggered.connect(partial(self._show_configuration,
                                              parent=parent_menu.parent()))
        config_menu.addAction(show_action)
        device = self.project_device
        disabled = device.status in NO_CONFIG_STATUSES or device.online
        for config in self.model.configs:
            conf_action = QAction(config.simple_name, config_menu)
            conf_action.setCheckable(True)
            callback = partial(self.active_config_changed, config)
            conf_action.triggered.connect(callback)
            is_active = self.model.active_config_ref == config.uuid
            conf_action.setChecked(is_active)
            config_menu.addAction(conf_action)
            conf_action.setEnabled(allowed and not disabled)

        return config_menu

    # ----------------------------------------------------------------------
    # QAction handlers

    def _show_device(self):
        deviceId = self.model.instance_id
        broadcast_event(KaraboEvent.ShowDevice,
                        {'deviceId': deviceId, 'showTopology': True})

    def _delete_device(self, project_controller, parent=None):
        """ Remove the device associated with this item from its device server
        """
        device = self.model
        development = get_config()["development"]
        if not development:
            ask = ('Are you sure you want to delete \"<b>{}</b>\".<br /> '
                   'Continue action?'.format(device.instance_id))
            msg_box = QMessageBox(QMessageBox.Question, 'Delete device',
                                  ask, QMessageBox.Yes | QMessageBox.No,
                                  parent=parent)
            msg_box.setModal(False)
            msg_box.setDefaultButton(QMessageBox.No)
            move_to_cursor(msg_box)
            if msg_box.exec() != QMessageBox.Yes:
                return

        server_model = find_parent_object(device, project_controller.model,
                                          DeviceServerModel)
        if device in server_model.devices:
            server_model.devices.remove(device)

    def _edit_device(self, project_controller, parent=None):
        # Watch for incomplete model initialization
        if not self.model.initialized:
            return

        device_model = self.model
        server_model = find_parent_object(
            device_model, project_controller.model, DeviceServerModel)
        dialog = DeviceHandleDialog(server_id=server_model.server_id,
                                    model=device_model,
                                    is_online=self.project_device.online,
                                    parent=parent)
        move_to_cursor(dialog)
        result = dialog.exec()
        if result == QDialog.Accepted:
            # Check for existing device
            new_name = device_model.instance_id != dialog.instance_id
            if new_name and check_device_instance_exists(dialog.instance_id):
                return

            # We might create a new project device here when renaming!
            device_model.instance_id = dialog.instance_id

            project_device = self.project_device
            new_class = device_model.class_id != dialog.class_id
            # Look for existing DeviceConfigurationModel
            config_model = device_model.select_config(dialog.active_uuid)
            assert config_model is not None, "Missing configuration"
            config_model.class_id = dialog.class_id

            if new_class:
                class_id = dialog.class_id
                # Note: Can only happen for offline project device
                assert not project_device.online
                # Set device model to take care of all configurations
                # and set new device
                device_model.class_id = class_id

            device_model.active_config_ref = dialog.active_uuid
            # When renaming, we have to apply the config hash!
            if new_name or new_class:
                self._update_icon(self.ui_data)
                self._update_instance_status(self.ui_data)
                config_hash = config_model.configuration
                project_device.set_project_config_hash(config_hash)
                self._update_configurator(True)

    def _about_device(self, parent=None):
        device = self.model
        info = {}
        info["deviceId"] = device.instance_id
        info["classId"] = device.class_id
        info["serverId"] = device.server_id
        info["uuid"] = device.uuid

        htmlString = ("<table>" +
                      "".join("<tr><td><b>{}</b>:   </td><td>{}</td></tr>".
                              format(*p) for p in info.items()) + "</table>")
        messagebox.show_information(htmlString, parent=parent)

    def _move_up(self, project_controller):
        device = self.model
        server_model = find_parent_object(device, project_controller.model,
                                          DeviceServerModel)

        if device in server_model.devices:
            index = server_model.devices.index(device)
            new_index = index - 1
            if new_index < 0:
                return
            devices = list(server_model.devices)
            del server_model.devices[:]
            devices.remove(device)
            devices.insert(new_index, device)
            server_model.devices.extend(devices)

    def _move_down(self, project_controller):
        device = self.model
        server_model = find_parent_object(device, project_controller.model,
                                          DeviceServerModel)

        if device in server_model.devices:
            index = server_model.devices.index(device)
            new_index = index + 1
            if new_index > len(server_model.devices):
                return
            devices = list(server_model.devices)
            del server_model.devices[:]
            devices.remove(device)
            devices.insert(new_index, device)
            server_model.devices.extend(devices)

    def _show_configuration(self, parent=None):
        config_model = self.active_config
        if config_model is None:
            return

        project_device = self.project_device
        allowed = access_role_allowed(AccessRole.PROJECT_EDIT)
        disabled = (project_device.status in NO_CONFIG_STATUSES
                    or project_device.online)
        editable = allowed and not disabled

        configuration = config_model.configuration
        dialog = DeviceConfigurationDialog(
            name=config_model.simple_name, configuration=configuration,
            editable=editable,
            project_device=project_device, parent=parent)
        move_to_cursor(dialog)
        if dialog.exec() == QDialog.Accepted:
            self.active_config.configuration = dialog.configuration
            project_device.set_project_config_hash(dialog.configuration)
            proxy = project_device.proxy
            if proxy.online:
                return
            # Show the device configuration if already showing
            # XXX: Check if this is needed
            self._update_configurator()

    def _duplicate_device(self, project_controller, parent=None):
        """ Duplicate the active device configuration of the model

        :param project: The parent project controller the model belongs to
        """
        device = self.model
        server_model = find_parent_object(device, project_controller.model,
                                          DeviceServerModel)
        active_config = self.active_config
        if active_config is None:
            return

        dialog = ObjectDuplicateDialog(device.instance_id, parent=parent)
        move_to_cursor(dialog)
        if dialog.exec() == QDialog.Accepted:
            xml = write_project_model(active_config)
            for simple_name in dialog.duplicate_names:
                # Check for existing device
                if check_device_instance_exists(simple_name):
                    continue
                dupe_dev_conf = read_project_model(StringIO(xml))
                dupe_dev_conf.reset_uuid()
                dev_inst = DeviceInstanceModel(
                    class_id=device.class_id,
                    instance_id=simple_name,
                    active_config_ref=dupe_dev_conf.uuid,
                    configs=[dupe_dev_conf]
                )
                # Set initialized and modified last
                dev_inst.initialized = dev_inst.modified = True
                server_model.devices.append(dev_inst)

    def _instantiate_device(self, project_controller, parent=None):
        server = find_parent_object(self.model, project_controller.model,
                                    DeviceServerModel)
        self.instantiate(server, parent)

    def _load_macro_from_device(self, project_controller, parent=None):
        """Request a scene directly from a device
        """
        dialog = DeviceCapabilityDialog(
            device_id=self.model.instance_id,
            capability=Capabilities.PROVIDES_MACROS,
            parent=parent)
        if dialog.exec() == QDialog.Accepted:
            device_id = dialog.device_id
            macro_name = dialog.capa_name
            project = project_controller.model
            project_macros = {s.simple_name for s in project.macros}

            if f'{device_id}-{macro_name}' in project_macros:
                msg = ('A macro with that name already exists in the selected '
                       'project.')
                messagebox.show_warning(msg, title='Cannot Load Macro')
                return

            get_macro_from_server(device_id, macro_name=macro_name,
                                  project=project)

    def _load_scene_from_device(self, project_controller, parent=None):
        """Request a scene directly from a device
        """
        dialog = DeviceCapabilityDialog(device_id=self.model.instance_id,
                                        parent=parent)
        if dialog.exec() == QDialog.Accepted:
            device_id = dialog.device_id
            scene_name = dialog.capa_name
            project = project_controller.model
            project_scenes = {s.simple_name for s in project.scenes}

            if f'{device_id}|{scene_name}' in project_scenes:
                msg = ('A scene with that name already exists in the selected '
                       'project.')
                messagebox.show_warning(msg, title='Cannot Load Scene')
                return

            get_scene_from_server(device_id, scene_name=scene_name,
                                  project=project)

    def _get_configuration_from_past(self, parent=None):
        """Request a configuration from the datalog reader
        """
        dialog = ConfigurationFromPastDialog(
            instance_id=self.model.instance_id, parent=parent)
        dialog.move(QCursor.pos())
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    def _get_configuration_from_name(self, parent=None):
        """Request a configuration from name from configuration manager"""
        device_id = self.model.instance_id
        dialog = InitConfigurationDialog(
            instance_id=device_id, parent=parent)
        dialog.move(QCursor.pos())
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    def _store_online_configuration(self, parent=None):
        """Set the online to offline configuration of this project device"""
        if not self.project_device.online:
            return

        device = self.project_device

        def _set_device_configuration():
            success, config = device.get_user_edited_config_hash_online()
            if success:
                title = f"Device configuration for {device.device_id}"
                old = self.active_config.configuration
                dialog = ConfigComparisonDialog(title, old, config,
                                                parent=parent)
                if dialog.exec() == QDialog.Accepted:
                    # Note: Make sure the active config is stored and
                    # apply the configuration to the online system!
                    self.active_config.configuration = config
                    device.set_project_config_hash_online(config)
            else:
                messagebox.show_error(
                    "Online to offline configuration currently not possible.",
                    parent=parent)
                return

        # Note: The device is online, but we require an offline schema
        # to compare for online to offline config extraction
        proxy = device._offline_proxy
        has_offline = len(proxy.binding.value) > 0
        if has_offline:
            _set_device_configuration()
        else:

            def _schema_handler():
                proxy.on_trait_change(_schema_handler, "schema_update",
                                      remove=True)
                _set_device_configuration()

            # There must be a class schema, since we are online. Hence,
            # we can safely wait here for the class schema of the server!
            proxy.on_trait_change(_schema_handler, "schema_update")
            proxy.ensure_class_schema()

    def instantiate(self, server, parent=None):
        """ Instantiate this device instance on the given `server`

        :param server: The server this device belongs to
        """
        if self.project_device.online:
            return

        device = self.project_device
        classId = device.class_id
        serverId = device.server_id
        if device.status in NO_CLASS_STATUSES:
            msg = (f"Unknown class schema for device plugin {classId} on "
                   f"server {serverId}. Please make sure to install the "
                   "device plugin first.")
            get_logger().warning(msg)
            return

        proxy = device.proxy
        deviceId = proxy.device_id
        get_manager().initDevice(serverId, classId, deviceId)

    def shutdown_device(self, show_confirm=True, parent=None):
        device = self.model
        get_manager().shutdownDevice(device.instance_id, show_confirm, parent)



#############################################################################
# Author: <martin.teichmann@xfel.eu>
# Created on March 19, 2014
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import QSize, Qt, Slot
from qtpy.QtGui import QIcon, QPainter, QPen, QPixmap
from qtpy.QtWidgets import QColorDialog, QComboBox, QDialog, QFormLayout

from .utils import get_dialog_ui


class PenDialog(QDialog):
    linecaps = {Qt.FlatCap: "butt", Qt.SquareCap: "square",
                Qt.RoundCap: "round"}
    linejoins = {Qt.SvgMiterJoin: "miter", Qt.MiterJoin: "miter",
                 Qt.BevelJoin: "bevel", Qt.RoundJoin: "round"}
    pen_styles = [Qt.SolidLine, Qt.DashLine, Qt.DotLine, Qt.DashDotLine,
                  Qt.DashDotDotLine]

    def __init__(self, pen, brush=None, parent=None):
        super().__init__(parent)
        uic.loadUi(get_dialog_ui('pendialog.ui'), self)
        self.pen = pen
        self.brush = brush

        self.set_color()

        self.slStrokeOpacity.setValue(pen.color().alpha())

        if pen.style() == Qt.NoPen:
            self.sbStrokeWidth.setValue(0)
        else:
            self.sbStrokeWidth.setValue(pen.width())

        self.wDashType = PenStyleComboBox()
        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.wDashType)
        self.wDashType.setPenStyle(self._get_style_from_pattern(self.pen))

        self.dsbDashOffset.setValue(self.pen.dashOffset())

        cap_button = getattr(self, self.linecaps[self.pen.capStyle()] + 'Cap')
        cap_button.setChecked(True)
        join_name = self.linejoins[self.pen.joinStyle()] + 'Join'
        join_button = getattr(self, join_name)
        join_button.setChecked(True)
        self.dsbStrokeMiterLimit.setValue(self.pen.miterLimit())

        self.setBrushWidgets()

    def _get_style_from_pattern(self, pen):
        """ The pen style for the given ``pen`` is returned. If it is
            ``Qt.CustomDashLine`` the dash pattern is compared.
        """
        if pen.style() == Qt.CustomDashLine:
            p = QPen()
            for style in self.pen_styles:
                p.setStyle(style)
                if p.dashPattern() == pen.dashPattern():
                    return style
        return pen.style()

    @Slot()
    def on_pbStrokeColor_clicked(self):
        color = QColorDialog.getColor(self.pen.color())
        if color.isValid():
            self.pen.setColor(color)
            self.set_color()

    @Slot()
    def on_pbFillColor_clicked(self):
        # The button is now only visible when `self.brush` is not None.
        # Just in case, do nothing.
        if self.brush is None:
            return

        color = QColorDialog.getColor(self.brush.color())
        if color.isValid():
            self.brush.setColor(color)
            self.set_color()

    def set_color(self):
        p = QPixmap(32, 16)
        p.fill(self.pen.color())
        self.pbStrokeColor.setIcon(QIcon(p))

        if self.brush is not None:
            p.fill(self.brush.color())
            self.pbFillColor.setIcon(QIcon(p))

    def setBrushWidgets(self):
        if self.brush is None:
            # Hide brush related widgets
            self.gbFill.setHidden(True)
            self.adjustSize()
        else:
            if self.brush.style() == Qt.SolidPattern:
                self.gbFill.setChecked(True)
            else:
                self.gbFill.setChecked(False)
            self.slFillOpacity.setValue(self.brush.color().alpha())

    def on_dialog_accepted(self):
        strokeColor = self.pen.color()
        strokeColor.setAlpha(self.slStrokeOpacity.value())
        self.pen.setColor(strokeColor)

        if self.sbStrokeWidth.value() == 0:
            self.pen.setStyle(Qt.NoPen)
        else:
            self.pen.setWidth(self.sbStrokeWidth.value())
            # Set style first!
            pen_style = self.wDashType.penStyle()
            self.pen.setStyle(pen_style)

            if pen_style is not Qt.SolidLine:
                self.pen.setDashOffset(self.dsbDashOffset.value())

        for k, v in self.linecaps.items():
            if getattr(self, v + 'Cap').isChecked():
                self.pen.setCapStyle(k)
        for k, v in self.linejoins.items():
            if getattr(self, v + 'Join').isChecked():
                self.pen.setJoinStyle(k)

        self.pen.setMiterLimit(self.dsbStrokeMiterLimit.value())
        if self.brush is not None:
            if not self.gbFill.isChecked():
                self.brush.setStyle(Qt.NoBrush)
            else:
                self.brush.setStyle(Qt.SolidPattern)
                fillColor = self.brush.color()
                fillColor.setAlpha(self.slFillOpacity.value())
                self.brush.setColor(fillColor)

    def exec(self):
        result = QDialog.exec(self)
        if result == QDialog.Accepted:
            self.on_dialog_accepted()
        return result


class PenStyleComboBox(QComboBox):
    styles = [(Qt.SolidLine, "Solid line"),
              (Qt.DashLine, "Dashed line"),
              (Qt.DotLine, "Dot line"),
              (Qt.DashDotLine, "Dash dot line"),
              (Qt.DashDotDotLine, "Dash dot dot line")]

    def __init__(self, parent=None):
        super().__init__(parent)

        self.setIconSize(QSize(32, 12))
        for s in self.styles:
            style = s[0]
            name = s[1]
            self.addItem(self.iconForPen(style), name, style)

    def penStyle(self):
        return self.styles[self.currentIndex()][0]

    def setPenStyle(self, style):
        id = self.findData(style)
        if id == -1:
            id = 0
        self.setCurrentIndex(id)

    def iconForPen(self, style):
        pix = QPixmap(self.iconSize())
        p = QPainter()
        pix.fill(Qt.transparent)

        p.begin(pix)
        pen = QPen(style)
        pen.setWidth(2)
        p.setPen(pen)

        mid = int(self.iconSize().height() / 2)
        p.drawLine(0, mid, int(self.iconSize().width()), mid)
        p.end()

        return QIcon(pix)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QEvent, QPoint, Qt, Slot
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QStyle

from karabogui.widgets.range_slider import RangeSlider, SliderHandle


def test_range_slider_widget(gui_app):
    """Test the basic slider range widget"""

    # Test slider goofy protection
    widget = RangeSlider()
    widget.initialize(0.2, 99.8)
    widget.show()

    assert widget.minimum() == 0
    assert widget.maximum() == 99

    widget.setValue(0.12, 47.5)
    assert widget.minimum() == 0
    assert widget.maximum() == 99
    assert widget.value() == (0, 47)

    widget.setMaximum(67.123)
    assert widget.maximum() == 67

    widget.setMinimum(1.56)
    assert widget.minimum() == 1
    assert widget.value() == (1, 47)
    widget.destroy()

    # Round trip test
    low = None
    high = None

    @Slot(int, int)
    def sliderMoved(low_value, high_value):
        nonlocal low, high
        low = low_value
        high = high_value

    widget = RangeSlider()
    widget.sliderMoved.connect(sliderMoved)
    widget.initialize(0, 100)
    widget.show()

    assert widget.minimum() == 0
    assert widget.maximum() == 100
    assert widget.sliderPosition() == (0, 100)
    assert widget.value() == (0, 100)

    widget.setMaximum(120)
    assert widget.sliderPosition() == (0, 100)
    widget.setMinimum(-20)
    assert widget.sliderPosition() == (0, 100)

    # Set extremes wrong
    widget.setMaximum(90)
    assert widget.sliderPosition() == (0, 90)
    widget.setMinimum(20)
    assert widget.sliderPosition() == (20, 90)

    widget.setSliderPosition(30, 40)
    assert widget.sliderPosition() == (30, 40)

    widget.setValue(40, 50)
    assert widget.sliderPosition() == (40, 50)

    move_event = QMouseEvent(
        QEvent.MouseMove,
        QPoint(20, 0),
        Qt.LeftButton,
        Qt.LeftButton,
        Qt.NoModifier)

    # We need to click before this is active
    widget.mouseMoveEvent(move_event)
    assert widget.sliderPosition() == (40, 50)

    assert widget.pressedHandle is None
    assert widget.pressedControl == QStyle.SC_None
    press_event = QMouseEvent(
        QEvent.MouseButtonPress,
        QPoint(80, 0),
        Qt.LeftButton,
        Qt.LeftButton,
        Qt.NoModifier)

    widget.mousePressEvent(press_event)
    assert widget.pressedControl == QStyle.SC_SliderHandle
    assert widget.pressedHandle == SliderHandle.HIGH

    move_event = QMouseEvent(
        QEvent.MouseMove,
        QPoint(35, 45),
        Qt.LeftButton,
        Qt.LeftButton,
        Qt.NoModifier)

    # Protection, moving from high to low
    widget.mouseMoveEvent(move_event)

    assert widget.sliderPosition() == (33, 33)
    assert widget.pressedHandle == SliderHandle.HIGH
    assert low == 33
    assert high == 33

    # Spawn a groove
    widget.setValue(30, 50)
    widget.update()

    # Release event triggers unclicked
    release_event = QMouseEvent(
        QEvent.MouseButtonRelease,
        QPoint(80, 0),
        Qt.LeftButton,
        Qt.LeftButton,
        Qt.NoModifier)
    widget.mouseReleaseEvent(release_event)
    assert widget.pressedControl == QStyle.SC_None
    assert widget.pressedHandle is None

    press_event = QMouseEvent(
        QEvent.MouseButtonPress,
        QPoint(30, 0),
        Qt.LeftButton,
        Qt.LeftButton,
        Qt.NoModifier)
    widget.mousePressEvent(press_event)
    assert widget.pressedControl == QStyle.SC_SliderHandle
    assert widget.pressedHandle == SliderHandle.LOW

    move_event = QMouseEvent(
        QEvent.MouseMove,
        QPoint(35, 45),
        Qt.LeftButton,
        Qt.LeftButton,
        Qt.NoModifier)

    # Protection, moving from low to high
    widget.mouseMoveEvent(move_event)
    assert widget.sliderPosition() == (33, 50)
    assert widget.pressedHandle == SliderHandle.LOW
    assert low == 33
    assert high == 50

    # Release event triggers unclicked
    release_event = QMouseEvent(
        QEvent.MouseButtonRelease,
        QPoint(80, 0),
        Qt.LeftButton,
        Qt.LeftButton,
        Qt.NoModifier)
    widget.mouseReleaseEvent(release_event)
    assert widget.pressedControl == QStyle.SC_None
    assert widget.pressedHandle is None

    widget.setValue(0, 100)
    # Limits
    assert widget.value() == (20, 90)

    # Test goofy values
    widget.initialize(-999999999999999999999999999999999, 0)
    assert not widget.isEnabled()
    assert not widget.isVisible()

    widget.setEnabled(True)
    widget.setVisible(True)
    widget.initialize(0, 9999999999999999999999999999999999999999)
    assert not widget.isEnabled()
    assert not widget.isVisible()

    widget.close()
    widget.destroy()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.

from pytestqt.modeltest import ModelTester

from karabo.common.project.api import (
    DeviceInstanceModel, DeviceServerModel, MacroModel, ProjectModel)
from karabo.common.scenemodel.api import SceneModel
from karabogui.project.view import ProjectView
from karabogui.singletons.project_model import ProjectViewItemModel
from karabogui.singletons.selection_tracker import SelectionTracker
from karabogui.testing import singletons


def test_basic_project(gui_app):
    """Test the QModelTest of the ProjectViewItemModel"""

    # 1. Macros
    macros = [MacroModel(simple_name="MyMacro", code="")]
    # 2. Scenes
    scenes = [SceneModel(simple_name="Empty Scene")]
    # 3. Devices
    device_karabo = DeviceInstanceModel(
        instance_id='Karabo')
    device_xfel = DeviceInstanceModel(
        instance_id='XFEL')
    # 4. Servers
    servers = [DeviceServerModel(devices=[device_karabo, device_xfel])]

    # 5. Nested subprojects
    subsubproject = ProjectModel(macros=macros)
    subprojects = [ProjectModel(macros=macros,
                                subprojects=[subsubproject])]
    # 6. Root model
    project_model = ProjectModel(
        macros=macros, scenes=scenes, servers=servers,
        subprojects=subprojects)
    tracker = SelectionTracker()
    with singletons(selection_tracker=tracker):
        model = ProjectViewItemModel()
        model.root_model = project_model
        with singletons(project_model=model):
            view = ProjectView()

        assert view.model() is model
        tester = ModelTester(None)
        tester.check(model)

        assert model.selectModel(device_karabo)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import GraphicsLayoutWidget

from karabogui.graph.common.api import AuxPlots
from karabogui.testing import GuiTestCase

from ..controller import AGGREGATOR_MAP, AuxPlotsController


class TestAuxPlots(GuiTestCase):

    def setUp(self):
        super().setUp()
        self._image_layout = GraphicsLayoutWidget()
        self._controller = AuxPlotsController(image_layout=self._image_layout)

    def test_basics(self):
        controllers = self._controller._aggregators
        assert self._controller.current_plot == AuxPlots.NoPlot
        assert controllers == {}

        # Add aux plot controllers
        for aux_plot, klass in AGGREGATOR_MAP.items():
            self._controller.add(aux_plot)
            assert aux_plot in controllers
            assert isinstance(controllers[aux_plot], klass)

        assert self._controller.current_plot == AuxPlots.NoPlot

        layout_plots = set(self._image_layout.ci.items.keys())
        assert len(layout_plots) == 0

        # Show profile plot first
        self._controller.current_plot = AuxPlots.ProfilePlot
        self._assert_plots(shown=AuxPlots.ProfilePlot)

        # Show histogram next
        self._controller.current_plot = AuxPlots.Histogram
        self._assert_plots(shown=AuxPlots.Histogram)

        # Don't show anything
        self._controller.current_plot = AuxPlots.NoPlot
        self._assert_plots(shown=AuxPlots.NoPlot)

        # Clear controller
        self._controller.destroy()
        assert len(self._image_layout.ci.items) == 0

    def _assert_plots(self, *, shown):
        layout_plots = self._image_layout.ci.items.keys()

        for plot_type, aggregator in self._controller._aggregators.items():
            assertion = (self.assertIn if plot_type == shown
                         else self.assertNotIn)
            for plot in aggregator.controllers.values():
                assertion(plot.plotItem, layout_plots)
